def import_sphinx():
    """
    sphinx is an optional extra, so only load it when necessary
    """
    try:
        import sphinx
    except ImportError:
        raise ImportError(
            "Sphinx is not installed. "
            "Please install ipypublish with the sphinx extras: "
            ">> pip install ipypublish[sphinx]"
        )
    if not sphinx.__version__ >= "1.8":
        raise ImportError("Sphinx version must be >= 1.8")
    return sphinx



"""
adapted from nbsphinx
"""
import re
import os

try:
    from urllib.parse import unquote  # Python 3.x
except ImportError:
    from urllib2 import unquote  # Python 2.x

import docutils
import nbconvert

from ipypublish.sphinx.utils import import_sphinx


class CreateNotebookSectionAnchors(docutils.transforms.Transform):
    """Create section anchors for Jupyter notebooks.

    Note: Sphinx lower-cases the HTML section IDs, Jupyter doesn't.
    This transform creates anchors in the Jupyter style.

    # TODO do this in pandoc filter?
    """

    default_priority = 200  # Before CreateSectionLabels (250)

    def apply(self):
        for section in self.document.traverse(docutils.nodes.section):
            title = section.children[0].astext()
            link_id = title.replace(" ", "-")
            section["ids"] = [link_id]


def _local_file_from_reference(node, document):
    """Get local file path from reference and split it into components."""
    # NB: Anonymous hyperlinks must be already resolved at this point!
    refuri = node.get("refuri")
    if not refuri:
        refname = node.get("refname")
        if refname:
            refid = document.nameids.get(refname)
        else:
            # NB: This can happen for anonymous hyperlinks
            refid = node.get("refid")
        target = document.ids.get(refid)
        if not target:
            # No corresponding target, Sphinx may warn later
            return "", "", ""
        refuri = target.get("refuri")
        if not refuri:
            # Target doesn't have URI
            return "", "", ""
    if "://" in refuri:
        # Not a local link
        return "", "", ""
    elif refuri.startswith("#") or refuri.startswith("mailto:"):
        # Not a local link
        return "", "", ""

    # NB: We look for "fragment identifier" before unquoting
    match = re.match(r"^([^#]+)(\.[^#]+)(#.+)$", refuri)
    if match:
        base = unquote(match.group(1))
        # NB: The suffix and "fragment identifier" are not unquoted
        suffix = match.group(2)
        fragment = match.group(3)
    else:
        base, suffix = os.path.splitext(refuri)
        base = unquote(base)
        fragment = ""
    return base, suffix, fragment


class RewriteLocalLinks(docutils.transforms.Transform):
    """Turn links to source files into ``:doc:``/``:ref:`` links.

    Links to subsections are possible with ``...#Subsection-Title``.
    These links use the labels created by CreateSectionLabels.

    Links to subsections use ``:ref:``, links to whole source files use
    ``:doc:``.  Latter can be useful if you have an ``index.rst`` but
    also want a distinct ``index.ipynb`` for use with Jupyter.
    In this case you can use such a link in a notebook::

        [Back to main page](index.ipynb)

    In Jupyter, this will create a "normal" link to ``index.ipynb``, but
    in the files generated by Sphinx, this will become a link to the
    main page created from ``index.rst``.

    """

    default_priority = 500  # After AnonymousHyperlinks (440)

    def apply(self):
        sphinx = import_sphinx()
        env = self.document.settings.env
        for node in self.document.traverse(docutils.nodes.reference):
            base, suffix, fragment = _local_file_from_reference(node, self.document)
            if not base:
                continue

            for s in env.config.source_suffix:
                if suffix.lower() == s.lower():
                    target = base
                    if fragment:
                        target_ext = suffix + fragment
                        reftype = "ref"
                    else:
                        target_ext = ""
                        reftype = "doc"
                    break
            else:
                continue  # Not a link to a potential Sphinx source file

            target_docname = nbconvert.filters.posix_path(
                os.path.normpath(os.path.join(os.path.dirname(env.docname), target))
            )
            if target_docname in env.found_docs:
                reftarget = "/" + target_docname + target_ext
                if reftype == "ref":
                    reftarget = reftarget.lower()
                linktext = node.astext()
                xref = sphinx.addnodes.pending_xref(
                    reftype=reftype,
                    reftarget=reftarget,
                    refdomain="std",
                    refwarn=True,
                    refexplicit=True,
                    refdoc=env.docname,
                )
                xref += docutils.nodes.Text(linktext, linktext)
                node.replace_self(xref)


class CreateSectionLabels(docutils.transforms.Transform):
    """Make labels for each document and each section thereof.

    These labels are referenced in RewriteLocalLinks but can also be
    used manually with ``:ref:``.

    """

    default_priority = 250  # Before references.PropagateTargets (260)

    def apply(self):
        env = self.document.settings.env
        file_ext = os.path.splitext(env.doc2path(env.docname))[1]
        i_still_have_to_create_the_document_label = True
        for section in self.document.traverse(docutils.nodes.section):
            assert section.children
            assert isinstance(section.children[0], docutils.nodes.title)
            title = section.children[0].astext()
            link_id = section["ids"][0]
            label = "/" + env.docname + file_ext + "#" + link_id
            label = label.lower()
            env.domaindata["std"]["labels"][label] = (env.docname, link_id, title)
            env.domaindata["std"]["anonlabels"][label] = (env.docname, link_id)

            # Create a label for the whole document using the first section:
            if i_still_have_to_create_the_document_label:
                label = "/" + env.docname.lower() + file_ext
                env.domaindata["std"]["labels"][label] = (env.docname, "", title)
                env.domaindata["std"]["anonlabels"][label] = (env.docname, "")
                i_still_have_to_create_the_document_label = False


class CreateDomainObjectLabels(docutils.transforms.Transform):
    """Create labels for domain-specific object signatures."""

    default_priority = 250  # About the same as CreateSectionLabels

    def apply(self):
        sphinx = import_sphinx()
        env = self.document.settings.env
        file_ext = os.path.splitext(env.doc2path(env.docname))[1]
        for sig in self.document.traverse(sphinx.addnodes.desc_signature):
            try:
                title = sig["ids"][0]
            except IndexError:
                # Object has same name as another, so skip it
                continue
            link_id = title.replace(" ", "-")
            sig["ids"] = [link_id]
            label = "/" + env.docname + file_ext + "#" + link_id
            label = label.lower()
            env.domaindata["std"]["labels"][label] = (env.docname, link_id, title)
            env.domaindata["std"]["anonlabels"][label] = (env.docname, link_id)



import os
import sys
import pytest

from ipypublish.convert.config_manager import iter_all_export_paths


@pytest.mark.parametrize("plugin_name,plugin_path", list(iter_all_export_paths()))
@pytest.mark.ipynb("basic_nb")
def test_basic_all_plugins(ipynb_app, plugin_name, plugin_path):

    if (
        plugin_name in ["sphinx_ipypublish_all.ext"] or "exec" in plugin_name
    ) and sys.version_info[0] < 3:
        # TODO this fails because the kernel is set as python3 in the notebook
        # could add a replacement variable e.g. ${pykernel}
        # and allow parsing of it to main.publish (default = "")
        return

    ipynb_app.run({"conversion": plugin_name})

    # no output file to compare
    if plugin_name in [
        "python_with_meta_stream",
        "sphinx_ipypublish_all.ext",
        "sphinx_ipypublish_all.ext.noexec",
    ]:
        return

    # test build exists only
    if plugin_name in ["sphinx_ipypublish_all.run", "sphinx_ipypublish_main.run"]:
        outpath = ipynb_app.converted_path.joinpath(
            "build", "html", os.path.splitext(ipynb_app.input_file.name)[0] + ".html"
        )
        assert outpath.exists()
        return

    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_equals_expected(plugin_name)



#!/usr/bin/env python
"""
Some setup for improved latex/pdf output

at top of workbook, use:

.. code:: python

    from ipynb_latex_setup import *

"""
from __future__ import division as _division

# Py2/Py3 compatibility
# =====================
from __future__ import print_function as _print_function

import warnings

warnings.warn(
    "this approach is now deprecated, "
    "instead please use\n`from ipypublish.scripts import nb_setup`\n"
    "then use the individual functions it provides to setup matplotlib, etc",
    DeprecationWarning,
    stacklevel=2,
)


# PYTHON
# =======

# IPYTHON
# =======
try:
    from IPython import get_ipython
    from IPython.display import Image, Latex
    from IPython.display import set_matplotlib_formats

    _ipy_present = True
except ImportError:
    _ipy_present = False
if _ipy_present:
    ipython = get_ipython()
    if ipython is not None:
        try:
            ipython.magic("config InlineBackend.figure_format = 'svg'")
            ipython.magic("matplotlib inline")
            set_matplotlib_formats("pdf", "svg")
        except Exception:
            pass

# NUMPY
# =====
try:
    import numpy as np
except ImportError:
    pass

# MATPLOTLIB
# ===========
try:
    import matplotlib as mpl

    _mpl_present = True
except ImportError:
    _mpl_present = False

if _mpl_present:
    mpl.rcParams["savefig.dpi"] = 75
    mpl.rcParams["figure.figsize"] = (7, 4)
    mpl.rcParams["figure.autolayout"] = False
    mpl.rcParams["axes.labelsize"] = 18
    mpl.rcParams["axes.titlesize"] = 20
    mpl.rcParams["font.size"] = 16
    mpl.rcParams["lines.linewidth"] = 2.0
    mpl.rcParams["lines.markersize"] = 8
    mpl.rcParams["legend.fontsize"] = 14
    mpl.rcParams["text.usetex"] = True
    mpl.rcParams["font.family"] = "serif"
    mpl.rcParams["font.serif"] = "cm"
    mpl.rcParams["text.latex.preamble"] = r"\usepackage{subdepth}, \usepackage{type1cm}"

# PANDAS
# ======
try:
    import pandas as pd

    _pandas_present = True
except ImportError:
    _pandas_present = False

if _pandas_present and ipython:
    pd.set_option("display.latex.repr", True)
    pd.set_option("display.latex.longtable", False)
    pd.set_option("display.latex.escape", True)

# SYMPY
# =====
try:
    import sympy as sym

    _sympy_present = True
except ImportError:
    _sympy_present = False
if _sympy_present:
    sym.init_printing(use_latex=True)

# IMAGE ARRANGEMENT with PIL
# ==========================
try:
    from PIL import Image as PImage

    _pil_present = True
except ImportError:
    _pil_present = False
if _pil_present:

    def create_test_image(size=(50, 50)):
        from io import BytesIO

        file = BytesIO()
        image = PImage.new("RGBA", size=size, color=(155, 0, 0))
        image.save(file, "png")
        file.name = "test.png"
        file.seek(0)
        return file

    def images_read(paths):
        """read a list of image paths to a list of PIL.IMAGE instances """
        return [PImage.open(i).convert("RGBA") for i in paths]

    def images_hconcat(images, width=700, height=700, gap=0, aspaths=True):
        """concatenate multiple images horizontally

        Parameters
        ----------
        images : list
            if aspaths=True, list of path strings, else list of PIL.Image instances
        width : int or list[int]
            maximum width of final image, or of individual images
        height : int or list[int]
            maximum height of final image, or of individual images
        gap : int
            size of space between images

        Returns
        -------
        image : PIL.Image

        Examples
        --------
        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_hconcat([img_path,img_path])
        >>> img.size
        (100, 50)

        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_hconcat([img_path,img_path],width=40,height=40)
        >>> img.size
        (40, 20)

        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_hconcat([img_path,img_path],width=[40,30])
        >>> img.size
        (70, 40)

        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_hconcat([img_path,img_path],gap=10)
        >>> img.size
        (110, 50)

        """
        images = images_read(images) if aspaths else images
        if not isinstance(width, list):
            widths = [width for _ in images]
        else:
            widths = width[:]
            width = sum(widths) + gap * (len(images) - 1)
        if not isinstance(height, list):
            heights = [height for _ in images]
        else:
            heights = height[:]
            height = sum(heights)
        for im, w, h in zip(images, widths, heights):
            im.thumbnail((w, h), PImage.ANTIALIAS)
        widths, heights = zip(*(i.size for i in images))
        total_width = sum(widths) + gap * (len(images) - 1)
        max_height = max(heights)
        new_im = PImage.new("RGBA", (total_width, max_height))
        x_offset = 0
        for im in images:
            new_im.paste(im, (x_offset, 0), mask=im)
            x_offset += im.size[0] + gap
        new_im.thumbnail((width, height), PImage.ANTIALIAS)
        return new_im

    def images_vconcat(images, width=700, height=700, gap=0, aspaths=True):
        """concatenate multiple images vertically

        Parameters
        ----------
        images : list
            if aspaths=True, list of path strings, else list of PIL.Image instances
        width : int or list[int]
            maximum width of final image, or of individual images
        height : int or list[int]
            maximum height of final image, or of individual images
        gap : int
            size of space between images

        Returns
        -------
        image : PIL.Image

        Examples
        --------
        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_vconcat([img_path,img_path])
        >>> img.size
        (50, 100)

        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_vconcat([img_path,img_path],width=40,height=40)
        >>> img.size
        (20, 40)

        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_vconcat([img_path,img_path],width=[40,30])
        >>> img.size
        (40, 70)

        >>> img_path = create_test_image(size=(50,50))
        >>> img = images_vconcat([img_path,img_path],gap=10)
        >>> img.size
        (50, 110)


        """
        images = images_read(images) if aspaths else images
        if not isinstance(width, list):
            widths = [width for _ in images]
        else:
            widths = width[:]
            width = sum(widths)
        if not isinstance(height, list):
            heights = [height for _ in images]
        else:
            heights = height[:]
            height = sum(heights) + gap * (len(images) - 1)
        for im, w, h in zip(images, widths, heights):
            im.thumbnail((w, h), PImage.ANTIALIAS)
        widths, heights = zip(*(i.size for i in images))
        max_width = max(widths)
        total_height = sum(heights) + gap * (len(images) - 1)
        new_im = PImage.new("RGBA", (max_width, total_height))
        y_offset = 0
        for im in images:
            new_im.paste(im, (0, y_offset), mask=im)
            y_offset += im.size[1] + gap
        new_im.thumbnail((width, height), PImage.ANTIALIAS)
        return new_im

    def images_gridconcat(
        pathslist, width=700, height=700, aspaths=True, hgap=0, vgap=0
    ):
        """concatenate multiple images in a grid

        Parameters
        ----------
        pathslist : list[list]
            if aspaths=True, list of path strings, else list of PIL.Image instances
            each sub list constitutes a row
        width : int
            maximum width of final image
        height : int
            maximum height of final image
        hgap : int
            size of horizontal space between images
        vgap : int
            size of vertical space between images

        Returns
        -------
        image : PIL.Image

        """
        himages = [
            images_hconcat(paths, gap=hgap, aspaths=aspaths) for paths in pathslist
        ]
        new_im = images_vconcat(himages, gap=vgap, aspaths=False)
        new_im.thumbnail((width, height), PImage.ANTIALIAS)
        return new_im


# JSONEXTENDED
# ==========================
try:
    from jsonextended import ejson, edict

    _jsonextended_present = True
except ImportError:
    _jsonextended_present = False
if _jsonextended_present:
    from jsonextended import plugins as eplugins

    # eplugins.load_builtin_plugins()



from collections import deque
import io
import logging

logger = logging.getLogger(__name__)


def import_texsoup():
    try:
        from TexSoup import TexSoup
        from TexSoup.utils import TokenWithPosition
        from TexSoup.data import RArg, OArg
    except ImportError:
        raise ImportError(
            "to parse tex files, TexSoup must be installed: \n"
            "pip install texsoup\n"
            "conda install -c conda-forge texsoup"
        )
    except SyntaxError:
        raise ImportError(
            "TexSoup package is broken on python 2.7, "
            "so can not be imported for tex parsing"
        )
    return {
        "TexSoup": TexSoup,
        "RArg": RArg,
        "OArg": OArg,
        "TokenWithPosition": TokenWithPosition,
    }


def _create_msg_error(msg, node=None, row=None):
    """create error message, optionally including TexNode and row"""
    text = msg.strip()
    if row is not None:
        text = "(row {}) ".format(row) + text
    if hasattr(node, "name"):
        text = text + ": {}".format(node.name)
    return text


def extract_required_val(rarg):
    """extract the value of a TexSoup RArg"""
    RArg = import_texsoup()["RArg"]
    if not isinstance(rarg, RArg):
        raise ValueError("expected {} to be a required argument".format(type(rarg)))
    return rarg.value


def _extract_parameters(texsoup_exprs):
    """extract the parameters from a TexSoup expression list"""
    RArg = import_texsoup()["RArg"]
    TokenWithPosition = import_texsoup()["TokenWithPosition"]
    expressions = deque(texsoup_exprs)
    param_name = None
    params = {}
    errors = []
    while expressions:
        expr = expressions.popleft()
        if isinstance(expr, TokenWithPosition):
            # TODO is this the best way to extract parameter name?
            param_name = expr.text.replace(",", "").replace("=", "").strip()
        elif isinstance(expr, RArg):
            if param_name is None:
                errors.append(
                    "expected expression "
                    "'{}' to precede a parameter name".format(expr)
                )
                break
            if param_name in params:
                errors.append("parameter '{}' already defined".format(param_name))
            else:
                params[param_name] = expr.value
            param_name = None
        else:
            errors.append(
                "expected expression '{}' ".format(expr)
                + "to be a parameter name or required argument"
            )
            break

    if param_name is not None:
        pass  # allowed since last expr may be new line
        # errors.append(
        #     "parameter '{}' is not assigned a value".format(param_name))

    return params, errors


def extract_parameters(argument):
    """extract parameters from a TexSoup OArg or Arg"""
    RArg = import_texsoup()["RArg"]
    OArg = import_texsoup()["OArg"]
    if not isinstance(argument, (OArg, RArg)):
        raise ValueError(
            "expected {} to be of type OArg or RArg".format(type(argument))
        )

    opt_params, errors = _extract_parameters(argument.exprs)

    return opt_params, errors


def create_newgloss_dict(gterm, row=None):
    """
    """
    arguments = list(gterm.args)
    fields = {}

    if len(arguments) != 2:
        msg = _create_msg_error(
            "could not parse newglossaryterm (arguments != 2)", gterm, row
        )
        raise IOError(msg)

    key = extract_required_val(arguments[0])

    params, errors = extract_parameters(arguments[1])

    for error in errors:
        msg = _create_msg_error(
            "error reading 'parameter' block: {}".format(error), gterm, row
        )
        raise IOError(msg)

    for param_name, param_value in params.items():

        if param_name in fields:
            raise IOError(
                "duplicate parameter '{0}' in key '{1}'".format(param_name, key)
            )

        fields[param_name] = param_value

    return key, fields


def create_newacronym_dict(acronym, row=None):
    """
    """
    OArg = import_texsoup()["OArg"]

    arguments = list(acronym.args)
    fields = {}

    if len(arguments) < 3:
        msg = _create_msg_error(
            "could not parse newacronym (too few arguments)", acronym, row
        )
        raise IOError(msg)
    if len(arguments) > 4:
        msg = _create_msg_error(
            "could not parse newacronym (too many arguments)", acronym, row
        )
        raise IOError(msg)

    key = extract_required_val(arguments[-3])
    abbreviation = extract_required_val(arguments[-2])
    name = extract_required_val(arguments[-1])

    if len(arguments) == 4:
        options = arguments[0]

        if not isinstance(options, OArg):
            msg = _create_msg_error(
                "expected first argument of newacronym to be 'optional", acronym, row
            )
            raise IOError(msg)

        opt_params, errors = extract_parameters(options)

        for error in errors:
            msg = _create_msg_error(
                "error reading newacronym 'optional' block: {}".format(error),
                acronym,
                row,
            )
            raise IOError(msg)

        for opt_name, opt_value in opt_params.items():
            if opt_name in fields:
                raise IOError(
                    "duplicate parameter '{0}' in key '{1}'".format(opt_name, key)
                )
            fields[opt_name] = opt_value

    return key, abbreviation, name, fields


def parse_tex(
    text_str=None,
    path=None,
    encoding="utf8",
    abbrev_field="abbreviation",
    fname_field="longname",
    skip_ioerrors=False,
):
    """parse a tex file containing newglossaryentry and/or newacronym to dict

    Parameters
    ----------
    text_str=None: str
        string representing the tex file
    path=None: str
        path to the tex file
    encoding='utf8': str
        tex file encoding
    abbrev_field="abbreviation": str
        field key for acronym abbreviation
    fname_field="longname": str
        field key for acronym full name
    skip_ioerrors=False: bool
        skip errors on reading a single entry

    Returns
    -------
    dict: glossaryterms
        {key: fields}
    dict: acronyms
        {key: fields}

    """
    TexSoup = import_texsoup()["TexSoup"]

    if sum([e is not None for e in [text_str, path]]) != 1:
        raise ValueError("only one of text_str or path must be supplied")
    elif path is not None:
        if text_str is not None:
            raise ValueError("text_str and path cannot be set at the same time")
        with io.open(path, encoding=encoding) as fobj:
            text_str = fobj.read()

    latex_tree = TexSoup(text_str)

    keys = []
    gterms = {}
    acronyms = {}

    for gterm in latex_tree.find_all("newglossaryentry"):
        try:
            key, fields = create_newgloss_dict(gterm)
        except IOError:
            if skip_ioerrors:
                continue
            raise
        if key in keys:
            raise KeyError("duplicate key: {}".format(key))
        keys.append(key)
        gterms[key] = fields

    for acronym in latex_tree.find_all("newacronym"):
        try:
            key, abbreviation, name, fields = create_newacronym_dict(acronym)
        except IOError:
            if skip_ioerrors:
                continue
            raise
        if key in keys:
            raise KeyError("duplicate key: {}".format(key))
        keys.append(key)
        fields[abbrev_field] = abbreviation
        fields[fname_field] = name
        acronyms[key] = fields

    return gterms, acronyms



import re
import sys
from textwrap import dedent
import pytest
from ipypublish.bib2glossary import BibGlossDB

bib_str = """\
    @glsterm{gtkey1,
    description = {the description},
    name = {name}
    }
    @glsterm{gtkey2,
    description = {the description of other},
    name = {other name}
    }
    @glsacronym{akey1,
    abbreviation = {ABRV},
    longname = {Abbreviation},
    description = {a description}
    }
    @glsacronym{akey2,
    abbreviation = {OTHER},
    longname = {Abbrev of other},
    plural = {OTHERs}
    }
    @glssymbol{skey1,
    description = {the description of symbol},
    name = {\\pi}
    }
    @badtype{bkey1,
    field = {text}
    }
    """

# TODO check for key duplication
# see https://github.com/sciunto-org/python-bibtexparser/issues/237

tex_str = """\
    \\newacronym[description={a description}]{akey1}{OTHER}{Abbreviation of other}
    \\newglossaryentry{gtkey1}{
        name={other name},
        description={the description of other}
    }
    \\newglossaryentry{skey1}{
        name={name},
        description={the description},
        type={symbols}
    }
    """  # noqa: E501


def test_load_bib_type_error():

    bibgloss = BibGlossDB()
    with pytest.raises(TypeError):
        bibgloss.load_bib(text_str=dedent(bib_str), ignore_nongloss_types=False)


def test_load_bib_type_ignore():

    bibgloss = BibGlossDB()
    bibgloss.load_bib(text_str=dedent(bib_str), ignore_nongloss_types=True)
    assert set(bibgloss.keys()) == {"gtkey1", "gtkey2", "akey1", "akey2", "skey1"}


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_load_tex():

    bibgloss = BibGlossDB()
    bibgloss.load_tex(text_str=dedent(tex_str))
    assert {k: e.type for k, e in bibgloss.items()} == {
        "gtkey1": "glsterm",
        "akey1": "glsacronym",
        "skey1": "glssymbol",
    }


def test_to_dict():
    bibgloss = BibGlossDB()
    bibgloss.load_bib(text_str=dedent(bib_str), ignore_nongloss_types=True)
    dct = bibgloss.to_dict()
    assert set(dct.keys()) == {"gtkey1", "gtkey2", "akey1", "akey2", "skey1"}


def test_to_bib_string():
    bibgloss = BibGlossDB()
    bibgloss.load_bib(text_str=dedent(bib_str), ignore_nongloss_types=True)
    string = bibgloss.to_bib_string()
    assert re.search(
        "@glsacronym\\{akey1,.*@glsterm\\{gtkey1,.*@glssymbol\\{skey1.*",
        string,
        re.DOTALL,
    )


def test_to_latex_dict():
    bibgloss = BibGlossDB()
    bibgloss.load_bib(text_str=dedent(bib_str), ignore_nongloss_types=True)
    latex_dict = bibgloss.to_latex_dict()
    print(latex_dict)
    assert latex_dict == {
        ("glsacronym", "akey1"): [
            ("\\newacronym[description={a description}]{" "akey1}{ABRV}{Abbreviation}")
        ],
        ("glsacronym", "akey2"): [
            ("\\newacronym[plural={OTHERs}]{" "akey2}{OTHER}{Abbrev of other}")
        ],
        ("glsterm", "gtkey1"): [
            "\\newglossaryentry{gtkey1}{",
            "    description={the description},",
            "    name={name}",
            "}",
        ],
        ("glsterm", "gtkey2"): [
            "\\newglossaryentry{gtkey2}{",
            "    description={the description of other},",
            "    name={other name}",
            "}",
        ],
        ("glssymbol", "skey1"): [
            "\\newglossaryentry{skey1}{",
            "    description={the description of symbol},",
            "    name={\\pi},",
            "    type={symbols}",
            "}",
        ],
    }



""" define the bibglossary directive

Originally adapted from: https://github.com/mcmtroffaes/sphinxcontrib-bibtex
"""

import ast  # parse(), used for filter
import os.path  # getmtime()
import sphinx.util

from docutils.parsers.rst import Directive, directives
from sphinx.util.console import bold, standout

from ipypublish.bib2glossary import BibGlossDB
from ipypublish.sphinx.gls.nodes import BibGlossaryNode
from ipypublish.sphinx.gls.cache import BibliographyCache, BibfileCache


logger = sphinx.util.logging.getLogger(__name__)


class BibGlossaryDirective(Directive):

    """Class for processing the ``bibglossary`` directive.

    Parses the bibliography files, and produces a
    :class:`~ipypublish.sphinx.gls.nodes.BibGlossaryNode` node.

    .. seealso::

       Further processing of the resulting
       :class:`~ipypublish.sphinx.gls.nodes.BibGlossaryNode`
       node is done by
       :class:`~ipypublish.sphinx.gls.transforms.BibGlossaryTransform`.
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = False
    option_spec = {
        "cited": directives.flag,
        "notcited": directives.flag,
        "all": directives.flag,
        "unsorted": directives.flag,
        "filter": directives.unchanged,
        "style": directives.unchanged,
        "encoding": directives.encoding,
        "keyprefix": directives.unchanged,
    }

    _allowed_styles = "list"
    _default_style = "list"

    def run(self):
        """Process .bib files, set file dependencies, and create a
        node that is to be transformed to the entries of the
        glossary.
        """
        env = self.state.document.settings.env
        # create id and cache for this node
        # this id will be stored with the node
        # and is used to look up additional data in env.bibgloss_cache
        # (implementation note: new_serialno only guarantees unique
        # ids within a single document, but we need the id to be
        # unique across all documents, so we also include the docname
        # in the id)
        id_ = "bibtex-bibglossary-%s-%s" % (env.docname, env.new_serialno("bibgloss"))

        # set filter option
        if "filter" in self.options:
            if "all" in self.options:
                logger.warning(standout(":filter: overrides :all:"))
            if "notcited" in self.options:
                logger.warning(standout(":filter: overrides :notcited:"))
            if "cited" in self.options:
                logger.warning(standout(":filter: overrides :cited:"))
            try:
                filter_ = ast.parse(self.options["filter"])
            except SyntaxError:
                logger.warning(
                    standout("syntax error in :filter: expression")
                    + " ("
                    + self.options["filter"]
                    + "); "
                    "the option will be ignored"
                )
                filter_ = ast.parse("cited")
        elif "all" in self.options:
            filter_ = ast.parse("True")
        elif "notcited" in self.options:
            filter_ = ast.parse("not cited")
        else:
            # the default filter: include only cited entries
            filter_ = ast.parse("cited")

        style = self.options.get("style", env.app.config.bibgloss_default_style)
        if style not in self._allowed_styles:
            logger.warning(
                "style '{}' not in allowed styles, defaulting to '{}'".format(
                    style, self._default_style
                )
            )
            style = self._default_style

        bibcache = BibliographyCache(
            style=style,
            unsorted=("unsorted" in self.options),
            filter_=filter_,
            encoding=self.options.get(
                "encoding", self.state.document.settings.input_encoding
            ),
            keyprefix=self.options.get("keyprefix", ""),
            labels={},
            plurals={},
            bibfiles=[],
        )

        for bibfile in self.arguments[0].split():
            # convert to normalized absolute path to ensure that the same file
            # only occurs once in the cache
            bibfile = os.path.normpath(env.relfn2path(bibfile.strip())[1])
            # if the bibfile has been supplied with no extension, guess path
            bibfile = BibGlossDB().guess_path(bibfile) or bibfile
            self.process_bibfile(bibfile, bibcache.encoding)
            env.note_dependency(bibfile)
            bibcache.bibfiles.append(bibfile)
        env.bibgloss_cache.set_bibliography_cache(env.docname, id_, bibcache)
        return [BibGlossaryNode("", ids=[id_])]

    def update_bibfile_cache(self, bibfile, mtime, encoding):
        """Parse *bibfile*,  and store the
        parsed data, along with modification time *mtime*, in the
        bibtex cache.

        Parameters
        ----------
        bibfile: str
            The bib file name.
        mtime: float
            The bib file's modification time.

        """
        logger.info(bold("parsing bibtex file {0}... ".format(bibfile)), nonl=True)
        bibglossdb = BibGlossDB()
        bibglossdb.load(path=bibfile, encoding=encoding)
        logger.info("parsed {0} entries".format(len(bibglossdb)))
        env = self.state.document.settings.env
        env.bibgloss_cache.bibfiles[bibfile] = BibfileCache(
            mtime=mtime, data=bibglossdb
        )

    def process_bibfile(self, bibfile, encoding):
        """Check if ``env.bibgloss_cache.bibfiles[bibfile]`` is still
        up to date. If not, parse the *bibfile*, and store parsed data in the
        bibtex cache.

        Parameters
        ----------
        bibfile: str
            The bib file name.

        """
        env = self.state.document.settings.env
        cache = env.bibgloss_cache.bibfiles
        # get modification time of bibfile
        try:
            mtime = os.path.getmtime(bibfile)
        except OSError:
            logger.warning(standout("could not open bibtex file {0}.".format(bibfile)))
            cache[bibfile] = BibfileCache(  # dummy cache
                mtime=-float("inf"), data=BibGlossDB()
            )
            return
        # get cache and check if it is still up to date
        # if it is not up to date, parse the bibtex file
        # and store it in the cache
        logger.info(
            bold("checking for {0} in bibtex cache... ".format(bibfile)), nonl=True
        )
        try:
            bibfile_cache = cache[bibfile]
        except KeyError:
            logger.info("not found")
            self.update_bibfile_cache(bibfile, mtime, encoding)
        else:
            if mtime != bibfile_cache.mtime:
                logger.info("out of date")
                self.update_bibfile_cache(bibfile, mtime, encoding)
            else:
                logger.info("up to date")



""" a panflute filter to prepare document labelling in markdown files:

1) Add a ``$$reference`` key to the Document metadata

Then, for each Image, Math and Table found;

2) Extract labels and attributes to the right of Math or Table captions,
in the form; ``{#id .class-name a="an attribute"}``

3) If attributes found, remove them from the document and wrap the associated
   Math/Table in a Span/Div with the attributes and an additional class:
   ``labelled-Math`` or ``labelled-Table``

4) For all labelled Tables, Math and Images,
   place in metadata as e.g.
   meta["$$references"][label] = {"type": "Math", "number": 1}

For example:

    '$$a=1$$ {#a b=$2$}'


would be converted to this html:

.. code-block:: html

   <p>
   <span id="a" class="labelled-Math" data-b="2">
   <span class="math inline"><em>a</em> = 1</span>
   </span>
   </p>

"""
from panflute import Element, Doc, Table, Inline  # noqa: F401
import panflute as pf

from ipypublish.filters_pandoc.utils import (
    compare_version,
    get_panflute_containers,
    find_attributes,
)

LABELLED_IMAGE_CLASS = "labelled-Image"
LABELLED_MATH_CLASS = "labelled-Math"
LABELLED_TABLE_CLASS = "labelled-Table"

REFTYPE_TABLE = "Table"
REFTYPE_IMAGE = "Image"
REFTYPE_MATH = "Math"


def resolve_tables(element, doc):
    # type: (Table, Doc) -> None
    if not isinstance(element, (pf.Table)):
        return None

    ref_type = REFTYPE_TABLE

    attributes = None
    if element.caption:  # type: Inline
        # attributes = _find_attribute(element.caption[0],
        #                              allow_any=True, delete_preceding=False)
        attributes = find_attributes(
            element.caption[-1], search_left=True, include_element=True
        )

    if not attributes:
        return None

    # update count
    doc.refcount[ref_type] += 1
    # add to metadata
    doc.metadata["$$references"][attributes["id"]] = pf.MetaMap(
        **{"type": pf.MetaString(ref_type), "number": doc.refcount[ref_type]}
    )
    # remove attribute from caption
    element.caption = [el for el in element.caption if el not in attributes["elements"]]

    # wrap in a div
    return pf.Div(
        element,
        classes=["labelled-{}".format(ref_type)] + attributes["classes"],
        attributes=attributes["attributes"],
        identifier=attributes["id"],
    )


def resolve_equations_images(element, doc):
    # type: (Element, Doc) -> None

    # attribute equations in table captions / definition items?
    if not isinstance(element, get_panflute_containers(pf.Math)):
        return None

    if not element.content:
        return None

    to_delete = set()
    to_wrap = dict()

    subel = element.content[0]

    while subel:  # type: Element

        ref_type = None
        if isinstance(subel, pf.Math):
            ref_type = REFTYPE_MATH
        # elif isinstance(subel, pf.Table):
        #     ref_type = "Table"
        elif isinstance(subel, pf.Image):
            ref_type = REFTYPE_IMAGE
        else:
            subel = subel.next
            continue

        if isinstance(subel, pf.Image) and compare_version("1.16", ">="):
            # pandoc >= 1.16 already supports this
            # TODO for pandoc < 1.16 also look for attributes attached,
            # to the image path, as occurs with image references
            # see https://github.com/tomduck/pandoc-fignos/issues/14
            attributes = {
                "id": subel.identifier,
                # "classes": subel.classes,
                # "attributes": subel.attributes,
                "elements": [],
            }

        else:
            attributes = find_attributes(subel)
            if attributes:
                to_wrap[subel] = attributes
                for _ in attributes["elements"]:
                    subel = subel.next

        if attributes and attributes["id"]:
            # update count
            doc.refcount[ref_type] += 1
            # add to metadata
            doc.metadata["$$references"][attributes["id"]] = pf.MetaMap(
                **{"type": pf.MetaString(ref_type), "number": doc.refcount[ref_type]}
            )

            to_delete.update(attributes["elements"])

        subel = subel.next

    new_content = [
        pf.Span(
            el,
            classes=["labelled-{}".format(ref_type)] + to_wrap[el]["classes"],
            attributes=to_wrap[el]["attributes"],
            identifier=to_wrap[el]["id"],
        )
        if el in to_wrap
        else el
        for el in element.content
        if el not in to_delete
    ]

    # if isinstance(element, pf.Plain):
    #     return pf.Plain(*new_content)
    # else:
    #     return pf.Para(*new_content)
    element.content = new_content
    return element


def prepare(doc):
    # type: (Doc) -> None
    doc.refcount = {"Table": 0, "Image": 0, "Math": 0}
    doc.metadata["$$references"] = pf.MetaMap()


def finalize(doc):
    # type: (Doc) -> None
    del doc.refcount


def main(doc=None):
    # type: (Doc) -> None
    return pf.run_filters(
        [resolve_tables, resolve_equations_images], prepare, finalize, doc=doc
    )


if __name__ == "__main__":
    main()



"""
pandoc filters for the ipypublish format
"""
# import sys
from six import string_types
import panflute as pf

pf.elements.RAW_FORMATS.add("latex")
pf.elements.RAW_FORMATS.add("tex")
pf.elements.RAW_FORMATS.add("rst")
pf.elements.RAW_FORMATS.add("html5")


# see https://github.com/sergiocorreia/panflute/issues/97
# TODO delete this when fixed
# if sys.version_info[0] == 2 or True:
def builtin2meta(val):
    if isinstance(val, bool):
        return pf.MetaBool(val)
    elif isinstance(val, (float, int)):
        return pf.MetaString(str(val))
    elif isinstance(val, string_types):
        return pf.MetaString(val)
    elif isinstance(val, list):
        return pf.MetaList(*[builtin2meta(x) for x in val])
    elif isinstance(val, dict):
        return pf.MetaMap(*[(k, builtin2meta(v)) for k, v in val.items()])
    elif isinstance(val, pf.Block):
        return pf.MetaBlocks(val)
    elif isinstance(val, pf.Inline):
        return pf.MetaInlines(val)
    elif isinstance(
        val,
        (
            pf.MetaBool,
            pf.MetaString,
            pf.MetaValue,
            pf.MetaList,
            pf.MetaMap,
            pf.MetaBlocks,
            pf.MetaInlines,
        ),
    ):
        return val

    raise TypeError("unknown type: {} (type: {})".format(val, type(val)))


pf.elements.builtin2meta = builtin2meta



"""
Some useful functions for creating
publishable jupyter notebooks

usage:

.. code-block:: python

    from ipypublish import nb_setup
    plt = nb_setup.setup_matplotlib(
        print_errors=True,
        output=('pdf',))
    pd = nb_setup.setup_pandas(escape_latex=True)
    sym = nb_setup.setup_sympy()
    import numpy as np
    from IPython.display import Image, Latex

"""
# Py2/Py3 compatibility
# =====================
# from __future__ import division as _division
# from __future__ import print_function as _print_function
import json
from io import BytesIO

# from IPython.display import Image, Latex

MPL_OPTIONS = (
    ("lines.linewidth", 1.5),
    ("lines.markeredgewidth", 1.0),
    ("lines.markersize", 8),
    ("lines.antialiased", True),
    ("lines.dashed_pattern", (3.7, 1.6)),
    ("lines.dashdot_pattern", (6.4, 1.6, 1, 1.6)),
    ("lines.dotted_pattern", [1, 1.65]),
    ("lines.scale_dashes", True),
    ("patch.linewidth", 1.0),
    ("patch.force_edgecolor", False),
    ("patch.antialiased", True),
    ("hatch.linewidth", 1.0),
    ("hist.bins", 10),
    ("boxplot.notch", False),
    ("boxplot.vertical", True),
    ("boxplot.whiskers", 1.5),
    ("boxplot.patchartist", False),
    ("boxplot.showmeans", False),
    ("boxplot.showcaps", True),
    ("boxplot.showbox", True),
    ("boxplot.showfliers", True),
    ("boxplot.meanline", False),
    ("boxplot.flierprops.markersize", 6),
    ("boxplot.flierprops.linewidth", 1.0),
    ("boxplot.boxprops.linewidth", 1.0),
    ("boxplot.whiskerprops.linewidth", 1.0),
    ("boxplot.capprops.linewidth", 1.0),
    ("boxplot.medianprops.linewidth", 1.0),
    ("boxplot.meanprops.markersize", 6),
    ("boxplot.meanprops.linewidth", 1.0),
    ("font.family", "sans-serif"),
    ("font.serif", "cm"),
    ("font.size", 16),
    ("text.usetex", True),
    # ('text.latex.unicode', False),
    ("text.latex.preamble", ("\\usepackage{subdepth}", "\\usepackage{type1cm}")),
    ("text.latex.preview", False),
    ("text.hinting_factor", 8),
    ("text.antialiased", True),
    ("mathtext.fallback_to_cm", True),
    ("image.lut", 256),
    ("image.resample", True),
    ("image.composite_image", True),
    ("contour.corner_mask", True),
    ("errorbar.capsize", 0),
    ("axes.labelsize", 18),
    ("axes.linewidth", 2.0),
    ("axes.spines.left", True),
    ("axes.spines.right", True),
    ("axes.spines.bottom", True),
    ("axes.spines.top", True),
    ("axes.titlesize", 20),
    ("axes.titlepad", 6.0),
    ("axes.grid", False),
    ("axes.labelpad", 4.0),
    ("axes.formatter.limits", (-7, 7)),
    ("axes.formatter.use_locale", False),
    ("axes.formatter.use_mathtext", False),
    ("axes.formatter.min_exponent", 0),
    ("axes.formatter.useoffset", True),
    ("axes.formatter.offset_threshold", 4),
    ("axes.unicode_minus", True),
    ("axes.xmargin", 0.05),
    ("axes.ymargin", 0.05),
    ("polaraxes.grid", True),
    ("axes3d.grid", True),
    ("legend.fontsize", 14),
    ("legend.fancybox", True),
    ("legend.numpoints", 1),
    ("legend.scatterpoints", 1),
    ("legend.markerscale", 1.0),
    ("legend.shadow", False),
    ("legend.frameon", True),
    ("legend.framealpha", 0.8),
    ("legend.borderpad", 0.4),
    ("legend.labelspacing", 0.5),
    ("legend.handlelength", 2.0),
    ("legend.handleheight", 0.7),
    ("legend.handletextpad", 0.8),
    ("legend.borderaxespad", 0.5),
    ("legend.columnspacing", 2.0),
    ("xtick.top", False),
    ("xtick.bottom", True),
    ("xtick.labeltop", False),
    ("xtick.labelbottom", True),
    ("xtick.major.size", 3.5),
    ("xtick.minor.size", 2),
    ("xtick.major.width", 0.8),
    ("xtick.minor.width", 0.6),
    ("xtick.major.pad", 3.5),
    ("xtick.minor.pad", 3.4),
    ("xtick.minor.visible", False),
    ("xtick.minor.top", True),
    ("xtick.minor.bottom", True),
    ("xtick.major.top", True),
    ("xtick.major.bottom", True),
    ("ytick.left", True),
    ("ytick.right", False),
    ("ytick.labelleft", True),
    ("ytick.labelright", False),
    ("ytick.major.size", 3.5),
    ("ytick.minor.size", 2),
    ("ytick.major.width", 0.8),
    ("ytick.minor.width", 0.6),
    ("ytick.major.pad", 3.5),
    ("ytick.minor.pad", 3.4),
    ("ytick.minor.visible", False),
    ("ytick.minor.left", True),
    ("ytick.minor.right", True),
    ("ytick.major.left", True),
    ("ytick.major.right", True),
    ("grid.linewidth", 0.8),
    ("grid.alpha", 1.0),
    ("figure.figsize", (6.4, 4.8)),
    ("figure.dpi", 100),
    ("figure.frameon", True),
    ("figure.autolayout", False),
    ("figure.max_open_warning", 20),
    ("figure.subplot.left", 0.125),
    ("figure.subplot.right", 0.9),
    ("figure.subplot.bottom", 0.11),
    ("figure.subplot.top", 0.88),
    ("figure.subplot.wspace", 0.2),
    ("figure.subplot.hspace", 0.2),
    ("figure.constrained_layout.use", False),
    ("figure.constrained_layout.hspace", 0.02),
    ("figure.constrained_layout.wspace", 0.02),
    ("figure.constrained_layout.h_pad", 0.04167),
    ("figure.constrained_layout.w_pad", 0.04167),
    ("savefig.jpeg_quality", 95),
    ("savefig.pad_inches", 0.1),
    ("savefig.transparent", False),
    ("savefig.dpi", 75),
    ("svg.image_inline", True),
    ("path.simplify", True),
    ("path.simplify_threshold", 0.1111111111111111),
    ("path.snap", True),
    ("path.effects", ()),
    ("agg.path.chunksize", 0),
    ("animation.embed_limit", 20),
    ("animation.bitrate", -1),
    ("animation.html_args", ()),
    ("animation.ffmpeg_args", ()),
    ("animation.avconv_args", ()),
    ("animation.convert_args", ()),
)


def setup_matplotlib(
    output=("pdf", "svg"), rcparams=None, usetex=True, print_errors=False
):
    """ import and setup matplotlib in the jupyter notebook

    Parameters
    ----------
    output: tuple[str]
        the output formats to save to the notebook
    rcparams: None or dict
        update default parameters set for matplotlib
    usetex: bool
        if True, and the 'latex' command is available,
        create figures with LaTeX
    print_errors: bool
        print errors for unavailable rcparams

    """
    from IPython import get_ipython
    from IPython.display import set_matplotlib_formats
    import matplotlib as mpl

    try:
        from shutil import which
    except ImportError:
        from shutilwhich import which

    ipython = get_ipython()
    latex_available = which("latex") is not None
    # if not latex_available:
    #     output = [o for o in output if o != "pdf"]

    set_matplotlib_formats(*output)
    ipython.magic("matplotlib inline")
    if "svg" in output:
        ipython.magic("config InlineBackend.figure_format = 'svg'")

    final_params = dict(MPL_OPTIONS)
    if rcparams is not None:
        final_params.update(rcparams)
    if usetex and latex_available:
        final_params.update({"text.usetex": True})
    else:
        final_params.update({"text.usetex": False})

    keyerrors = []
    valerrors = {}
    for key, val in final_params.items():
        try:
            mpl.rcParams[key] = val
        except KeyError:
            keyerrors.append(key)
        except ValueError:
            valerrors[key] = val

    if print_errors:
        if keyerrors:
            print("KeyErrors:")
            for key in keyerrors:
                print("- key")
        if valerrors:
            print("ValueError:")
            print(json.dumps(valerrors, indent=2))

    return mpl.pyplot


def setup_pandas(escape_latex=False, use_longtable=False):
    """ import and setup pandas in the jupyter notebook

    Parameters
    ----------
    escape_latex: bool
        whether to escape special latex character, e.g. `_` -> `\\_`

    """
    import pandas as pd

    pd.set_option("display.latex.repr", True)
    pd.set_option("display.latex.longtable", use_longtable)
    pd.set_option("display.latex.escape", escape_latex)
    return pd


def setup_sympy():
    import sympy

    sympy.init_printing(use_latex=True)
    return sympy


def get_pimage():
    try:
        from PIL import Image
    except ImportError:
        raise ImportError("to use this function; pip install pillow")
    return Image


def create_test_image(size=(50, 50)):
    file = BytesIO()
    image = get_pimage().new("RGBA", size=size, color=(155, 0, 0))
    image.save(file, "png")
    file.name = "test.png"
    file.seek(0)
    return file


def images_read(paths):
    """read a list of image paths to a list of PIL.IMAGE instances """
    return [get_pimage().open(i).convert("RGBA") for i in paths]


def images_hconcat(images, width=700, height=700, gap=0, aspaths=True):
    """concatenate multiple images horizontally

    Parameters
    ----------
    images : list
        if aspaths=True, list of path strings, else list of PIL.Image instances
    width : int or list[int]
        maximum width of final image, or of individual images
    height : int or list[int]
        maximum height of final image, or of individual images
    gap : int
        size of space between images

    Returns
    -------
    image : PIL.Image

    Examples
    --------
    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_hconcat([img_path,img_path])
    >>> img.size
    (100, 50)

    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_hconcat([img_path,img_path],width=40,height=40)
    >>> img.size
    (40, 20)

    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_hconcat([img_path,img_path],width=[40,30])
    >>> img.size
    (70, 40)

    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_hconcat([img_path,img_path],gap=10)
    >>> img.size
    (110, 50)

    """
    pimage = get_pimage()

    images = images_read(images) if aspaths else images
    if not isinstance(width, list):
        widths = [width for _ in images]
    else:
        widths = width[:]
        width = sum(widths) + gap * (len(images) - 1)
    if not isinstance(height, list):
        heights = [height for _ in images]
    else:
        heights = height[:]
        height = sum(heights)
    for im, w, h in zip(images, widths, heights):
        im.thumbnail((w, h), pimage.ANTIALIAS)
    widths, heights = zip(*(i.size for i in images))
    total_width = sum(widths) + gap * (len(images) - 1)
    max_height = max(heights)
    new_im = pimage.new("RGBA", (total_width, max_height))
    x_offset = 0
    for im in images:
        new_im.paste(im, (x_offset, 0), mask=im)
        x_offset += im.size[0] + gap
    new_im.thumbnail((width, height), pimage.ANTIALIAS)
    return new_im


def images_vconcat(images, width=700, height=700, gap=0, aspaths=True):
    """concatenate multiple images vertically

    Parameters
    ----------
    images : list
        if aspaths=True, list of path strings, else list of PIL.Image instances
    width : int or list[int]
        maximum width of final image, or of individual images
    height : int or list[int]
        maximum height of final image, or of individual images
    gap : int
        size of space between images

    Returns
    -------
    image : PIL.Image

    Examples
    --------
    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_vconcat([img_path,img_path])
    >>> img.size
    (50, 100)

    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_vconcat([img_path,img_path],width=40,height=40)
    >>> img.size
    (20, 40)

    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_vconcat([img_path,img_path],width=[40,30])
    >>> img.size
    (40, 70)

    >>> img_path = create_test_image(size=(50,50))
    >>> img = images_vconcat([img_path,img_path],gap=10)
    >>> img.size
    (50, 110)


    """
    pimage = get_pimage()

    images = images_read(images) if aspaths else images
    if not isinstance(width, list):
        widths = [width for _ in images]
    else:
        widths = width[:]
        width = sum(widths)
    if not isinstance(height, list):
        heights = [height for _ in images]
    else:
        heights = height[:]
        height = sum(heights) + gap * (len(images) - 1)
    for im, w, h in zip(images, widths, heights):
        im.thumbnail((w, h), pimage.ANTIALIAS)
    widths, heights = zip(*(i.size for i in images))
    max_width = max(widths)
    total_height = sum(heights) + gap * (len(images) - 1)
    new_im = pimage.new("RGBA", (max_width, total_height))
    y_offset = 0
    for im in images:
        new_im.paste(im, (0, y_offset), mask=im)
        y_offset += im.size[1] + gap
    new_im.thumbnail((width, height), pimage.ANTIALIAS)
    return new_im


def images_gridconcat(pathslist, width=700, height=700, aspaths=True, hgap=0, vgap=0):
    """concatenate multiple images in a grid

    Parameters
    ----------
    pathslist : list[list]
        if aspaths=True, list of path strings, else list of PIL.Image instances
        each sub list constitutes a row
    width : int
        maximum width of final image
    height : int
        maximum height of final image
    hgap : int
        size of horizontal space between images
    vgap : int
        size of vertical space between images

    Returns
    -------
    image : PIL.Image

    """
    pimage = get_pimage()
    himages = [images_hconcat(paths, gap=hgap, aspaths=aspaths) for paths in pathslist]
    new_im = images_vconcat(himages, gap=vgap, aspaths=False)
    new_im.thumbnail((width, height), pimage.ANTIALIAS)
    return new_im


if __name__ == "__main__":
    import matplotlib as mpl

    print(
        json.dumps(
            {
                k: v[0]
                for k, v in mpl.rcsetup.defaultParams.items()
                if isinstance(v[0], (int, float, list, tuple, set))
            },
            indent=2,
        )
    )



import copy
import logging

import traitlets as traits
from nbconvert.preprocessors import Preprocessor
from nbformat.notebooknode import NotebookNode


def merge(a, b, path=None, overwrite=True):
    """merges b into a

    Examples
    --------
    >>> from pprint import pprint
    >>> pprint(merge({'a':{'b':1},'c':3},{'a':{'b':2}}))
    {'a': {'b': 2}, 'c': 3}

    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)], overwrite)
            elif a[key] == b[key]:
                pass  # same leaf value
            elif overwrite:
                a[key] = b[key]  # overwrite leaf value
            else:
                raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


class SplitOutputs(Preprocessor):
    """ a preprocessor to split outputs into separate cells,
    merging the cell and output metadata, with output metadata taking priority

    """

    split = traits.Bool(True, help="whether to split outputs").tag(config=True)

    def preprocess(self, nb, resources):

        if not self.split:
            return nb, resources

        logging.info("splitting outputs into separate cells")

        final_cells = []
        for cell in nb.cells:

            if not cell.cell_type == "code":
                final_cells.append(cell)
                continue
            outputs = cell.pop("outputs")
            cell.outputs = []
            final_cells.append(cell)
            for output in outputs:
                meta = copy.deepcopy(cell.metadata)
                # don't need the code to output
                meta.get("ipub", NotebookNode({})).code = False
                # don't create a new slide for each output,
                # unless specified in output level metadata
                if "slide" in meta.get("ipub", NotebookNode({})):
                    if meta.ipub.slide == "new":
                        meta.ipub.slide = True
                    else:
                        meta.ipub.slide = meta.ipub.slide
                meta = merge(meta, output.get("metadata", {}))
                new = NotebookNode(
                    {
                        "cell_type": "code",
                        "source": "",
                        "execution_count": None,
                        "metadata": meta,
                        "outputs": [output],
                    }
                )
                final_cells.append(new)

        nb.cells = final_cells

        return nb, resources



#!/usr/bin/env python
import logging
import os
import sys
from mimetypes import guess_type

from ipypublish.frontend.shared import parse_options
from ipypublish.convert.main import IpyPubMain
from ipypublish.postprocessors.reveal_serve import RevealServer

logger = logging.getLogger("nbpresent")


def nbpresent(
    inpath,
    outformat="slides_standard",
    outpath=None,
    dump_files=True,
    ignore_prefix="_",
    clear_files=False,
    log_level="INFO",
    dry_run=False,
    print_traceback=False,
    export_paths=(),
):
    """ load reveal.js slides as a web server,
    converting from ipynb first if path extension is .ipynb

    Parameters
    ----------
    inpath: str
        path to html or ipynb file
    outformat: str
        conversion format to use
    outpath : str  or pathlib.Path
        path to output converted files
    dump_files: bool
        whether to write files from nbconvert (images, etc) to outpath
    clear_files : str
        whether to clear existing external files in outpath folder
    ignore_prefix: str
        ignore ipynb files with this prefix
    log_level: str
        the logging level (debug, info, critical, ...)

    """
    inpath_name, inpath_ext = os.path.splitext(os.path.basename(inpath))

    output_mimetype = guess_type(inpath, strict=False)[0]
    output_mimetype = "unknown" if output_mimetype is None else output_mimetype

    if output_mimetype != "text/html":

        config = {
            "IpyPubMain": {
                "conversion": outformat,
                "plugin_folder_paths": export_paths,
                "outpath": outpath,
                "ignore_prefix": ignore_prefix,
                "log_to_stdout": True,
                "log_level_stdout": log_level,
                "log_to_file": True,
                "log_level_file": log_level,
                "default_pporder_kwargs": dict(
                    dry_run=dry_run,
                    clear_existing=clear_files,
                    dump_files=dump_files,
                    serve_html=True,
                    slides=True,
                ),
            }
        }
        publish = IpyPubMain(config=config)
        try:
            outdata = publish(inpath)

            outpath = outdata["outpath"]
            output_mimetype = outdata["exporter"].output_mimetype

        except Exception as err:
            logger.error("Run Failed: {}".format(err))
            if print_traceback:
                raise
            return 1
    else:
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
        server = RevealServer()
        if not dry_run:
            server.postprocess("", output_mimetype, os.path.abspath(inpath))

    return 0


def run(sys_args=None):

    if sys_args is None:
        sys_args = sys.argv[1:]

    filepath, options = parse_options(sys_args, "nbpresent")

    outcode = nbpresent(filepath, **options)

    return outcode



#!/usr/bin/env python
# Note, updated version of
# https://github.com/ipython/ipython-in-depth/blob/master/tools/nbmerge.py
"""
usage:

python nbmerge.py directory_name > merged.ipynb
"""
from __future__ import print_function

import logging
import os
import re
import sys

import nbformat
from six import string_types

from ipypublish.utils import pathlib, handle_error

logger = logging.getLogger("nbmerge")


def alphanumeric_sort(l):
    """sort key.name alphanumerically

    Parameters
    ----------
    l: list[str]

    """

    def convert(text):
        return int(text) if text.isdigit() else text.lower()

    def alphanum_key(key):
        return [convert(c) for c in re.split("([0-9]+)", key.name)]

    return sorted(l, key=alphanum_key)


def merge_notebooks(ipynb_path, ignore_prefix="_", to_str=False, as_version=4):
    """ merge one or more ipynb's,
    if more than one, then the meta data is taken from the first

    Parameters
    ----------
    ipynb_path: str or pathlib.Path
    ignore_prefix : str
        ignore filename starting with this prefix
    to_str: bool
        return as a string, else return nbformat object
    as_version: int
        notebook format vesion

    Returns
    ------
    finalnb: jupyter.notebook
    meta_path : pathlib.Path
        path to notebook containing meta file

    """
    meta_path = ""
    if isinstance(ipynb_path, string_types):
        ipynb_path = pathlib.Path(ipynb_path)
    if not ipynb_path.exists():
        handle_error(
            "the notebook path does not exist: {}".format(ipynb_path), IOError, logger
        )

    final_nb = None
    if ipynb_path.is_dir():
        logger.info("Merging all notebooks in directory")
        for ipath in alphanumeric_sort(ipynb_path.glob("*.ipynb")):
            if os.path.basename(ipath.name).startswith(ignore_prefix):
                continue
            with ipath.open("r", encoding="utf-8") as f:
                if (
                    sys.version_info.major == 3
                    and sys.version_info.minor < 6
                    and "win" not in sys.platform
                ):
                    data = f.read()
                    if hasattr(data, "decode"):
                        data = data.decode("utf-8")
                    nb = nbformat.reads(data, as_version=as_version)
                else:
                    nb = nbformat.read(f, as_version=as_version)
            if final_nb is None:
                meta_path = ipath
                final_nb = nb
            else:
                final_nb.cells.extend(nb.cells)
    else:
        logger.info("Reading notebook")
        with ipynb_path.open("r", encoding="utf-8") as f:
            if (
                sys.version_info.major == 3
                and sys.version_info.minor < 6
                and "win" not in sys.platform
            ):
                data = f.read()
                if hasattr(data, "decode"):
                    data = data.decode("utf-8")
                final_nb = nbformat.reads(data, as_version=as_version)
            else:
                final_nb = nbformat.read(f, as_version=as_version)
        meta_path = ipynb_path
    if not hasattr(final_nb.metadata, "name"):
        final_nb.metadata.name = ""
    final_nb.metadata.name += "_merged"

    if to_str:
        if sys.version_info > (3, 0):
            return nbformat.writes(final_nb)
        else:
            return nbformat.writes(final_nb).encode("utf-8")

    if final_nb is None:
        handle_error(
            "no acceptable notebooks found for path: {}".format(ipynb_path.name),
            IOError,
            logger,
        )

    return final_nb, meta_path



import docutils

from ipypublish.bib2glossary import BibGlossEntry


def docutils_citation_ref_node(
    entry, document, use_key_as_label=True, classes=("bibglossary",)
):
    # type: (dict, docutils.nodes.document, bool) -> docutils.nodes.citation_reference  # noqa
    """Return citation_reference node to the given citation. The
    citation_reference is expected to be inserted into *document*
    prior to any docutils transforms.
    """
    if not isinstance(entry, BibGlossEntry):
        raise TypeError
    # see docutils.parsers.rst.states.Body.footnote_reference()
    if use_key_as_label:
        label = entry.key
    else:
        label = entry.label
    refname = docutils.nodes.fully_normalize_name(entry.key)
    refnode = docutils.nodes.citation_reference(
        rawsource="[%s]_" % label, refname=refname
    )
    label_text = docutils.nodes.Text(label)
    refnode += label_text
    refnode["classes"].extend(classes)
    document.note_citation_ref(refnode)
    return refnode


def docutils_citation_node(entry, document, use_key_as_label=True, convert_latex=True):
    """Return citation node, with key as name, label as first child,
    and nodes with entry text (converted from latex) as subsequent children.

    The citation is expected to be inserted into the *document*
    prior to any docutils transforms.
    """
    # see docutils.parsers.rst.states.Body.citation()
    if not isinstance(entry, BibGlossEntry):
        raise TypeError

    if use_key_as_label:
        label = entry.key
    else:
        label = entry.label
    name = docutils.nodes.fully_normalize_name(entry.key)

    citation = docutils.nodes.citation()
    citation["names"].append(name)
    citation += docutils.nodes.label("", label)
    if convert_latex:
        for child in latex_to_docutils(entry.text):
            citation += child
    else:
        citation += docutils.nodes.paragraph(
            "", "", docutils.nodes.Text(entry.text, entry.text)
        )
    citation["classes"].append("bibglossary")
    document.note_citation(citation)
    document.note_explicit_target(citation, citation)
    return citation


def format_entries(entries, style="list", sort=True):
    # TODO apply styles consistent with latex glossaries
    if sort:
        entries = sorted(entries, key=lambda e: e.sortkey)
    labels = [e.label for e in entries]
    for label, entry in zip(labels, entries):
        yield entry


def rst_to_docutils(source):
    parser = docutils.parsers.rst.Parser()
    settings = docutils.frontend.OptionParser(
        components=(docutils.parsers.rst.Parser,)
    ).get_default_values()
    document = docutils.utils.new_document("dummy_source_path", settings)
    parser.parse(source, document)
    return document


def latex_to_docutils(source):
    import panflute as pf

    rst_source = pf.convert_text(source, input_format="latex", output_format="rst")
    return rst_to_docutils(rst_source)



import os
import json
from ipypublish.port_api.plugin_to_json import convert_to_json
from ipypublish.port_api.tpl_dct_to_json import py_to_json
from ipypublish.port_api.convert_format_str import convert_format_str
from ipypublish.tests import TEST_FILES_DIR


def test_plugin_to_json_html():

    out_str = convert_to_json(
        os.path.join(TEST_FILES_DIR, "port_api_files", "html_ipypublish_all.py.txt")
    )
    dct = json.loads(out_str)
    assert "exporter" in dct  # TODO test against schema


def test_plugin_to_json_latex():

    out_str = convert_to_json(
        os.path.join(TEST_FILES_DIR, "port_api_files", "latex_ipypublish_all.py.txt")
    )
    dct = json.loads(out_str)
    assert "exporter" in dct  # TODO test against schema


def test_py_to_json():

    out_str = py_to_json(
        os.path.join(TEST_FILES_DIR, "port_api_files", "front_pages.py.txt")
    )
    dct = json.loads(out_str)
    assert "segments" in dct  # TODO test against schema


def test_convert_format_str():

    template = [
        "{{%- extends 'null.tpl' -%}}",
        "{{% block header %}}",
        "{{{{ nb.metadata | meta2yaml('#~~ ') }}}}",
        "{{% endblock header %}}",
        "{{% block codecell %}}",
        "#%%",
        "{{{{ super() }}}}",
        "{{% endblock codecell %}}",
        "{{% block in_prompt %}}{{% endblock in_prompt %}}",
        "{{% block input %}}{{{{ cell.metadata | meta2yaml('#~~ ') }}}}",  # noqa: E501
        "{{{{ cell.source | ipython2python }}}}",
        "{{% endblock input %}}",
        "{{% block markdowncell scoped %}}#%% [markdown]",
        "{{{{ cell.metadata | meta2yaml('#~~ ') }}}}",
        "{{{{ cell.source | comment_lines }}}}",
        "{{% endblock markdowncell %}}",
    ]

    convert_format_str(template)



from ipypublish.filters_pandoc.utils import apply_filter
from ipypublish.filters_pandoc.prepare_cites import main
from ipypublish.tests.utils import HTML2JSONParser


def test_para_rst():
    """
    """
    in_string = [
        "+@label{.class a=1} xyz *@label2* @label3{.b}",
        "",
        "(@label4{})",
        "",
        "(@label5{.b} x)",
    ]
    out_string = apply_filter(in_string, main, "rst")

    assert out_string.strip() == "\n".join(
        ["@label xyz *@label2* @label3", "", "(@label4)", "", "(@label5 x)"]
    )


def test_para_html(data_regression):
    """
    """
    in_string = ["+@label{ .class a=1} xyz *@label2* @label3{ .b}"]
    out_string = apply_filter(in_string, main, "html")

    parser = HTML2JSONParser()
    parser.feed(out_string)
    data_regression.check(parser.parsed)

    # assert out_string.strip() == "\n".join([
    #     '<p>'
    #     '<span class="attribute-Cite class" '
    #     'data-latex="cref" data-rst="numref" data-a="1">'
    #     '<span class="citation" data-cites="label">@label</span>'
    #     '</span> '
    #     'xyz '
    #     '<em><span class="citation" data-cites="label2">@label2</span></em> '
    #     '<span class="attribute-Cite b">'
    #     '<span class="citation" data-cites="label3">@label3</span>'
    #     '</span>'
    #     '</p>',
    # ])


def test_table_html(data_regression):
    """
    """
    in_string = ["a b", "- -", "x y", "", "Table: Caption +@label"]
    out_string = apply_filter(in_string, main, "html")

    parser = HTML2JSONParser()
    parser.feed(out_string)
    data_regression.check(parser.parsed)

    # assert out_string.strip() == "\n".join([
    #      '<table>',
    #      '<caption>Caption '
    #      '<span class="attribute-Cite" data-latex="cref" data-rst="numref">'
    #      '<span class="citation" data-cites="label">@label</span>'
    #      '</span></caption>',
    #      '<thead>',
    #      '<tr class="header">',
    #      '<th>a</th>',
    #      '<th>b</th>',
    #      '</tr>',
    #      '</thead>',
    #      '<tbody>',
    #      '<tr class="odd">',
    #      '<td>x</td>',
    #      '<td>y</td>',
    #      '</tr>',
    #      '</tbody>',
    #      '</table>'
    # ])


def test_image_html(data_regression):

    in_string = ["![a title with a @label1 +@label2 {.nclass x=3}](path/to/image.png)"]
    out_string = apply_filter(in_string, main, "html")

    parser = HTML2JSONParser()
    parser.feed(out_string)
    data_regression.check(parser.parsed)

    # assert out_string.strip() == "\n".join([
    #     '<figure>',
    #     '<img src="path/to/image.png" alt="a title with a @label1 @label2" />'
    #     '<figcaption>a title with a '
    #     '<span class="citation" data-cites="label1">@label1</span> '
    #     '<span class="attribute-Cite nclass" '
    #     'data-latex="cref" data-rst="numref" data-x="3">'
    #     '<span class="citation" data-cites="label2">@label2</span>'
    #     '</span></figcaption>',
    #     '</figure>'
    #     ])



import logging

import traitlets as traits
from nbconvert.preprocessors import Preprocessor
from nbformat.notebooknode import NotebookNode


def flatten(d, key_as_tuple=True, sep="."):
    """ get nested dict as {key:val,...},
    where key is tuple/string of all nested keys

    Parameters
    ----------
    d : dict
    key_as_tuple : bool
        whether keys are list of nested keys or delimited string of nested keys
    sep : str
        if key_as_tuple=False, delimiter for keys

    Examples
    --------

    >>> from pprint import pprint

    >>> d = {1:{"a":"A"},2:{"b":"B"}}
    >>> pprint(flatten(d))
    {(1, 'a'): 'A', (2, 'b'): 'B'}

    >>> d = {1:{"a":"A"},2:{"b":"B"}}
    >>> pprint(flatten(d,key_as_tuple=False))
    {'1.a': 'A', '2.b': 'B'}

    """

    def expand(key, value):
        if isinstance(value, dict):
            if key_as_tuple:
                return [(key + k, v) for k, v in flatten(value, key_as_tuple).items()]
            else:
                return [
                    (str(key) + sep + k, v)
                    for k, v in flatten(value, key_as_tuple).items()
                ]
        else:
            return [(key, value)]

    if key_as_tuple:
        items = [item for k, v in d.items() for item in expand((k,), v)]
    else:
        items = [item for k, v in d.items() for item in expand(k, v)]

    return dict(items)


class MetaDefaults(Preprocessor):
    """ a preprocessor which enters default metadata tags
    into all cell metadata, without overriding any currently set

    """

    nb_defaults = traits.Dict(
        default_value={}, help="dict of notebook level defaults"
    ).tag(config=True)
    cell_defaults = traits.Dict(
        default_value={}, help="dict of cell level defaults"
    ).tag(config=True)
    overwrite = traits.Bool(
        False, help="whether existing values should be overwritten"
    ).tag(config=True)

    def preprocess(self, nb, resources):

        logging.info("adding ipub defaults to notebook")

        for keys, val in flatten(self.nb_defaults).items():
            dct = nb.metadata
            for key in keys[:-1]:
                if key not in dct:
                    dct[key] = NotebookNode({})
                dct = dct[key]
            if keys[-1] not in dct:
                dct[keys[-1]] = val
            elif self.overwrite:
                dct[keys[-1]] = val

        for cell in nb.cells:

            for keys, val in flatten(self.cell_defaults).items():
                dct = cell.metadata
                leaf_not_dict = False
                for key in keys[:-1]:
                    if key not in dct:
                        dct[key] = NotebookNode({})
                    elif dct[key] is False and self.overwrite:
                        dct[key] = NotebookNode({})
                    elif dct[key] is True:
                        dct[key] = NotebookNode({})
                    elif not hasattr(dct[key], "items"):
                        leaf_not_dict = True
                        break
                    dct = dct[key]

                if leaf_not_dict:
                    pass
                elif keys[-1] not in dct:
                    dct[keys[-1]] = val
                elif self.overwrite:
                    dct[keys[-1]] = val

        return nb, resources



#!/usr/bin/env python
""" a module for exporting latex file to pdf
TODO could this be refactored as nbconvert postprocessor
"""
import os
import shutil
import tempfile
from subprocess import Popen, PIPE, STDOUT
import webbrowser

import six
from traitlets import Bool, Unicode

from ipypublish.postprocessors.base import IPyPostProcessor


class PDFExport(IPyPostProcessor):
    """ a post processor to convert tex to pdf using latexmk
    """

    @property
    def allowed_mimetypes(self):
        return "text/latex"

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "pdf-export"

    files_folder = Unicode(
        "_static",
        help="the path (relative to the main file path) " "containing external files",
    ).tag(config=True)

    convert_in_temp = Bool(
        False,
        help="run conversion in a temporary directory, " "and copy back only PDF file",
    ).tag(config=True)

    debug_mode = Bool(False, help="run in debug mode").tag(config=True)

    open_in_browser = Bool(
        False, help="launch a html page containing a pdf browser"
    ).tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):
        """ should not be called directly

        Parameters
        ----------
        stream: str
            the main file contents
        filepath: None or pathlib.Path
            the path to the output file

        Returns
        -------
        stream: str
        filepath: None or pathlib.Path

        """
        self.logger.info("running pdf conversion")
        self._export_pdf(filepath)
        return stream, filepath, resources

    def _export_pdf(self, texpath):

        if not texpath.exists():
            self.handle_error(
                "the target file path does not exist: {}".format(texpath), IOError
            )

        texname = os.path.splitext(texpath.name)[0]
        # NOTE outdir was originally passed, but would this ever be different
        # to the texpath's parent

        external_files = texpath.parent.joinpath(self.files_folder)

        if external_files.exists() and not external_files.is_dir():
            self.handle_error(
                "the external folder path is not a directory: {}".format(
                    external_files
                ),
                IOError,
            )

        self.check_exe_exists(
            "latexmk",
            "requires the latexmk executable to run. "
            "See http://mg.readthedocs.io/latexmk.html#installation",
        )

        if self.convert_in_temp:
            out_folder = tempfile.mkdtemp()
            try:
                exitcode = self._run_latexmk(texpath, out_folder, external_files)
                if exitcode == 0:
                    shutil.copyfile(
                        os.path.join(out_folder, texname + ".pdf"),
                        str(texpath.parent.joinpath(texname + ".pdf")),
                    )
            finally:
                shutil.rmtree(out_folder)
        else:
            exitcode = self._run_latexmk(texpath, str(texpath.parent), external_files)

        if exitcode == 0:
            self.logger.info("pdf conversion complete")

            view_pdf = VIEW_PDF.format(pdf_name=texname.replace(" ", "%20") + ".pdf")
            view_pdf_path = texpath.parent.joinpath(texname + ".view_pdf.html")
            with view_pdf_path.open("w", encoding="utf-8") as fobj:
                fobj.write(six.u(view_pdf))
        else:
            self.handle_error(
                "pdf conversion failed: " "Try running with pdf-debug flag",
                RuntimeError,
            )

        if self.open_in_browser:
            #  2 opens the url in a new tab
            webbrowser.open(view_pdf_path.as_uri(), new=2)

        return

    def _run_latexmk(self, texpath, out_folder, external_files):
        """ run latexmk conversion
        """
        # make sure tex file in right place
        outpath = os.path.join(out_folder, texpath.name)
        if os.path.dirname(str(texpath)) != str(out_folder):
            self.logger.debug(
                "copying tex file to: {}".format(
                    os.path.join(str(out_folder), texpath.name)
                )
            )
            shutil.copyfile(str(texpath), os.path.join(str(out_folder), texpath.name))

        # make sure the external files folder is in right place
        if external_files.exists():
            self.logger.debug("external files folder set")
            outfilespath = os.path.join(out_folder, str(external_files.name))
            if str(external_files) != str(outfilespath):
                self.logger.debug("copying external files to: {}".format(outfilespath))
                if os.path.exists(outfilespath):
                    shutil.rmtree(outfilespath)
                shutil.copytree(str(external_files), str(outfilespath))

        # run latexmk in correct folder
        with change_dir(out_folder):
            latexmk = ["latexmk", "-xelatex", "-bibtex", "-pdf"]
            latexmk += [] if self.debug_mode else ["--interaction=batchmode"]
            latexmk += [outpath]
            self.logger.info("running: " + " ".join(latexmk))

            def log_latexmk_output(pipe):
                for line in iter(pipe.readline, b""):
                    self.logger.info("latexmk: {}".format(line.decode("utf-8").strip()))

            process = Popen(latexmk, stdout=PIPE, stderr=STDOUT)
            with process.stdout:
                log_latexmk_output(process.stdout)
            exitcode = process.wait()  # 0 means success

        return exitcode


class change_dir:  # noqa: N801
    """Context manager for changing the current working directory"""

    def __init__(self, new_path):
        self.newPath = os.path.expanduser(new_path)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


VIEW_PDF = r"""
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">

<head>
    <meta http-equiv="content-type" content="text/html; charset=windows-1252">
    <title>View PDF</title>

    <script type="text/javascript">
       var filepath = "{pdf_name}";
       var timer = null;

       function refresh(){{
          var d = document.getElementById("pdf"); // gets pdf-div
          d.innerHTML = '<iframe style="position: absolute; height: 100%; border: none" id="ipdf" src='+window.filepath+'  width="100%"></iframe>';
       }}

       function autoRefresh(){{
          timer = setTimeout("autoRefresh()", 20000);
          refresh();
       }}

       function manualRefresh(){{
          clearTimeout(timer);
          refresh();
       }}
       function check_pdf() {{
         var newfile = document.f.userFile.value;
         ext = newfile.substring(newfile.length-3,newfile.length);
         ext = ext.toLowerCase();
         if(ext != 'pdf') {{
           alert('You selected a .'+ext+
                 ' file; please select a .pdf file instead!'+filepath);
           return false; }}
         else
             alert(newfile);
             window.filepath = newfile;
             alert(filepath);
             refresh();
           return true; }}
    </script>

</head>
<body>
    <!-- <form name=f onsubmit="return check_pdf();"
        action='' method='POST' enctype='multipart/form-data'>
        <input type='submit' name='upload_btn' value='upload'>
        <input type='file' name='userFile' accept="application/pdf">
    </form> -->
    <button onclick="manualRefresh()">manual refresh</button>
    <button onclick="autoRefresh()">auto refresh</button>
    <div id="pdf"></div>
</body>
<script type="text/javascript">refresh();</script>
</html>
"""  # noqa: E501



import pytest
from ipypublish.filters_pandoc.utils import apply_filter
from ipypublish.filters_pandoc import prepare_labels
from ipypublish.filters_pandoc import format_label_elements


def test_math_span_latex():

    in_json = {
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {
                        "t": "Span",
                        "c": [
                            ["a", ["labelled-Math"], [["b", "2"]]],
                            [{"t": "Math", "c": [{"t": "DisplayMath"}, "a=1"]}],
                        ],
                    }
                ],
            }
        ],
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {
            "$$references": {
                "t": "MetaMap",
                "c": {
                    "a": {
                        "t": "MetaMap",
                        "c": {
                            "type": {"t": "MetaString", "c": "Math"},
                            "number": {"t": "MetaString", "c": "1"},
                        },
                    }
                },
            }
        },
    }

    out_string = apply_filter(
        in_json, format_label_elements.main, "latex", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [r"\begin{equation}a=1\label{a}\end{equation}"]
    )


def test_math_span_rst():

    in_json = {
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {
                        "t": "Span",
                        "c": [
                            ["a", ["labelled-Math"], [["b", "2"]]],
                            [{"t": "Math", "c": [{"t": "DisplayMath"}, "a=1"]}],
                        ],
                    }
                ],
            }
        ],
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {
            "$$references": {
                "t": "MetaMap",
                "c": {
                    "a": {
                        "t": "MetaMap",
                        "c": {
                            "type": {"t": "MetaString", "c": "Math"},
                            "number": {"t": "MetaString", "c": "1"},
                        },
                    }
                },
            }
        },
    }

    out_string = apply_filter(
        in_json, format_label_elements.main, "rst", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [
            ".. math::",
            "   :nowrap:",
            "   :label: a",
            "",
            r"   \begin{equation}a=1\end{equation}",
        ]
    )


@pytest.mark.skip(
    reason="there's an issue with pandoc outputting unicode in '/em> = 1'"
)
def test_math_span_html():

    in_json = {
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {
                        "t": "Span",
                        "c": [
                            ["a", ["labelled-Math"], [["b", "2"]]],
                            [{"t": "Math", "c": [{"t": "DisplayMath"}, "a=1"]}],
                        ],
                    }
                ],
            }
        ],
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {
            "$$references": {
                "t": "MetaMap",
                "c": {
                    "a": {
                        "t": "MetaMap",
                        "c": {
                            "type": {"t": "MetaString", "c": "Math"},
                            "number": {"t": "MetaString", "c": "1"},
                        },
                    }
                },
            }
        },
    }

    out_string = apply_filter(
        in_json, format_label_elements.main, "html", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [
            '<p><a id="a" class="anchor-link" name="#a">'
            "<br />"
            '<span class="math display"><em>a</em> = 1</span>'
            "<br />"
            "</a></p>"
        ]
    )


def test_math_md_to_rst():

    in_str = [
        "$$a = b$$ {#eq:id1}",
        "$$c &= d \\\\ other &= e$$ {#a env=align .unnumbered}",
    ]
    out_string = apply_filter(
        in_str,
        [prepare_labels.main, format_label_elements.main],
        in_format="markdown",
        out_format="rst",
    )

    assert out_string.strip() == "\n".join(
        [
            ".. math::",
            "   :nowrap:",
            "   :label: eq:id1",
            "",
            r"   \begin{equation}a = b\end{equation}",
            "",
            "",
            "",
            ".. math::",
            "   :nowrap:",
            "   :label: a",
            "",
            r"   \begin{align*}c &= d \\ other &= e\end{align*}",
        ]
    )


def test_image_html():
    """
    """
    #  "![a title](path/to/image.png){#label1 .class-name a=5}"
    in_json = {
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {
                        "t": "Image",
                        "c": [
                            ["label1", ["class-name"], [["a", "5"]]],
                            [
                                {"t": "Str", "c": "a"},
                                {"t": "Space"},
                                {"t": "Str", "c": "title"},
                            ],
                            ["path/to/image.png", "fig:"],
                        ],
                    }
                ],
            }
        ],
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {},
    }

    out_string = apply_filter(
        in_json, format_label_elements.main, "html", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [
            '<p><a id="label1" class="anchor-link" name="#label1">'
            '<img src="path/to/image.png" title="fig:" alt="a title" id="label1" '
            'class="class-name" data-a="5" />'
            "</a></p>"
        ]
    )


def test_image_rst():
    """
    """
    #  "![a title](path/to/image.png){#label1 .class-name a=5}"
    in_json = {
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {
                        "t": "Image",
                        "c": [
                            ["label1", ["class-name"], [["a", "5"]]],
                            [
                                {"t": "Str", "c": "a"},
                                {"t": "Space"},
                                {"t": "Str", "c": "title"},
                            ],
                            ["path/to/image.png", "fig:"],
                        ],
                    }
                ],
            }
        ],
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {},
    }

    out_string = apply_filter(
        in_json, format_label_elements.main, "rst", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [
            ".. figure:: path/to/image.png",
            "   :alt: a title",
            "   :figclass: class-name",
            "   :name: label1",
            "",
            "   a title",
        ]
    )


def test_image_latex():
    """
    """
    #  "![a title](path/to/image.png){#label1 .class-name a=5}"
    in_json = {
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {
                        "t": "Image",
                        "c": [
                            ["label1", ["class-name"], [["a", "5"]]],
                            [
                                {"t": "Str", "c": "a"},
                                {"t": "Space"},
                                {"t": "Str", "c": "title"},
                            ],
                            ["path/to/image.png", "fig:"],
                        ],
                    }
                ],
            }
        ],
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {},
    }

    out_string = apply_filter(
        in_json, format_label_elements.main, "latex", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [
            r"\begin{figure}[]",
            r"\hypertarget{label1}{%",
            r"\begin{center}",
            r"\adjustimage{max size={0.9\linewidth}{0.9\paperheight},}"
            r"{path/to/image.png}",
            r"\end{center}",
            r"\caption{a title}\label{label1}",
            "}",
            r"\end{figure}",
        ]
    )


def test_table_html():
    """
    Some text

    a b
    - -
    1 2
    4 5

    Table: Caption. {#tbl:id}
    """
    in_json = {
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {
            "$$references": {
                "t": "MetaMap",
                "c": {
                    "tbl:id": {
                        "t": "MetaMap",
                        "c": {
                            "type": {"t": "MetaString", "c": "Table"},
                            "number": {"t": "MetaString", "c": "1"},
                        },
                    }
                },
            }
        },
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {"t": "Str", "c": "Some"},
                    {"t": "Space"},
                    {"t": "Str", "c": "text"},
                ],
            },
            {
                "t": "Div",
                "c": [
                    ["tbl:id", ["labelled-Table"], []],
                    [
                        {
                            "t": "Table",
                            "c": [
                                [{"t": "Str", "c": "Caption."}, {"t": "Space"}],
                                [{"t": "AlignDefault"}, {"t": "AlignDefault"}],
                                [0, 0],
                                [
                                    [{"t": "Plain", "c": [{"t": "Str", "c": "a"}]}],
                                    [{"t": "Plain", "c": [{"t": "Str", "c": "b"}]}],
                                ],
                                [
                                    [
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "1"}]}],
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "2"}]}],
                                    ],
                                    [
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "4"}]}],
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "5"}]}],
                                    ],
                                ],
                            ],
                        }
                    ],
                ],
            },
        ],
    }
    out_string = apply_filter(
        in_json, format_label_elements.main, "html", in_format="json"
    )

    assert out_string.strip() == "\n".join(
        [
            "<p>Some text</p>",
            '<a id="tbl:id" class="anchor-link" name="#tbl:id">',
            "<table>",
            "<caption>Caption. </caption>",
            "<thead>",
            '<tr class="header">',
            "<th>a</th>",
            "<th>b</th>",
            "</tr>",
            "</thead>",
            "<tbody>",
            '<tr class="odd">',
            "<td>1</td>",
            "<td>2</td>",
            "</tr>",
            '<tr class="even">',
            "<td>4</td>",
            "<td>5</td>",
            "</tr>",
            "</tbody>",
            "</table>",
            "</a>",
        ]
    )


def test_table_rst():
    """
    Some text

    a b
    - -
    1 2
    4 5

    Table: Caption. {#tbl:id}
    """
    in_json = {
        "pandoc-api-version": [1, 17, 5, 1],
        "meta": {
            "$$references": {
                "t": "MetaMap",
                "c": {
                    "tbl:id": {
                        "t": "MetaMap",
                        "c": {
                            "type": {"t": "MetaString", "c": "Table"},
                            "number": {"t": "MetaString", "c": "1"},
                        },
                    }
                },
            }
        },
        "blocks": [
            {
                "t": "Para",
                "c": [
                    {"t": "Str", "c": "Some"},
                    {"t": "Space"},
                    {"t": "Str", "c": "text"},
                ],
            },
            {
                "t": "Div",
                "c": [
                    ["tbl:id", ["labelled-Table"], []],
                    [
                        {
                            "t": "Table",
                            "c": [
                                [{"t": "Str", "c": "Caption."}, {"t": "Space"}],
                                [{"t": "AlignDefault"}, {"t": "AlignDefault"}],
                                [0, 0],
                                [
                                    [{"t": "Plain", "c": [{"t": "Str", "c": "a"}]}],
                                    [{"t": "Plain", "c": [{"t": "Str", "c": "b"}]}],
                                ],
                                [
                                    [
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "1"}]}],
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "2"}]}],
                                    ],
                                    [
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "4"}]}],
                                        [{"t": "Plain", "c": [{"t": "Str", "c": "5"}]}],
                                    ],
                                ],
                            ],
                        }
                    ],
                ],
            },
        ],
    }
    out_string = apply_filter(
        in_json, format_label_elements.main, "rst", in_format="json"
    )

    assert out_string.strip().splitlines()[0:3] == ["Some text", "", ".. _`tbl:id`:"]



import webbrowser
import shutil
import os
from subprocess import Popen, PIPE, STDOUT
from distutils.spawn import find_executable

from six import u
from traitlets import TraitError, validate, Bool, Dict, Unicode

from ipypublish import __version__
from ipypublish.postprocessors.base import IPyPostProcessor
from ipypublish.sphinx.utils import import_sphinx
from ipypublish.sphinx.create_setup import make_conf, make_index


# NOTE Interesting note about adding a directive to actually run python code
# https://stackoverflow.com/questions/7250659/how-to-use-python-to-programmatically-generate-part-of-sphinx-documentation


class RunSphinx(IPyPostProcessor):
    """ run sphinx to create an html output
    """

    @property
    def allowed_mimetypes(self):
        return ("text/restructuredtext",)

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "run-sphinx"

    open_in_browser = Bool(
        True, help="launch a html page containing a pdf browser"
    ).tag(config=True)

    numbered = Bool(
        True, help="set :numbered: in toc, which numbers sections, etc"
    ).tag(config=True)

    show_prompts = Bool(True, help="whether to include cell prompts").tag(config=True)

    prompt_style = Unicode("[{count}]:", help="the style of cell prompts").tag(
        config=True
    )

    @validate("prompt_style")
    def _valid_prompt_style(self, proposal):
        try:
            proposal.format(count=1)
        except TypeError:
            raise TraitError(
                "prompt_style should be formatable by " "`prompt_style.format(count=1)`"
            )
        return proposal["value"]

    conf_kwargs = Dict(
        help=(
            "additional key-word arguments to be included in the conf.py "
            "as <key> = <val>"
        )
    ).tag(config=True)

    override_defaults = Bool(
        True, help="if True, conf_kwargs override default values"
    ).tag(config=True)

    nitpick = Bool(False, help="nit-picky mode, warn about all missing references")

    def run_postprocess(self, stream, mimetype, filepath, resources):

        # check sphinx is available and the correct version
        try:
            import_sphinx()
        except ImportError as err:
            self.handle_error(err.msg, ImportError)

        self.logger.info("Creating Sphinx files")

        titlepage = {}
        if "titlepage" in resources.get("ipub", {}):
            titlepage = resources["ipub"]["titlepage"]
        #  includes ['author', 'email', 'supervisors', 'title', 'subtitle',
        #            'tagline', 'institution', 'logo']

        # create a conf.py
        kwargs = {} if not self.conf_kwargs else self.conf_kwargs
        kwargs["ipysphinx_show_prompts"] = self.show_prompts
        kwargs["ipysphinx_input_prompt"] = self.prompt_style
        kwargs["ipysphinx_output_prompt"] = self.prompt_style
        if "author" in titlepage:
            kwargs["author"] = titlepage["author"]
        if "tagline" in titlepage:
            kwargs["description"] = titlepage["tagline"]
        if "email" in titlepage:
            kwargs["email"] = titlepage["email"]

        conf_str = make_conf(overwrite=self.override_defaults, **kwargs)
        conf_path = filepath.parent.joinpath("conf.py")
        with conf_path.open("w", encoding="utf8") as f:
            f.write(u(conf_str))

        # create an index.rst
        toc_files = [filepath.name]

        toc_depth = 3
        title = None
        prolog = []
        epilog = []
        if "author" in titlepage:
            prolog.append(".. sectionauthor:: {0}".format(titlepage["author"]))
            prolog.append("")
        if "title" in titlepage:
            title = titlepage["title"]
        if "tagline" in titlepage:
            prolog.append(titlepage["tagline"])
        if "institution" in titlepage:
            for inst in titlepage["institution"]:
                epilog.append("| " + inst)

        epilog.append("")
        epilog.append("Created by IPyPublish (version {})".format(__version__))

        toc = resources.get("ipub", {}).get("toc", {})
        if hasattr(toc, "get") and "depth" in toc:
            toc_depth = toc["depth"]

        index_str = make_index(
            toc_files,
            toc_depth=toc_depth,
            header=title,
            toc_numbered=self.numbered,
            prolog="\n".join(prolog),
            epilog="\n".join(epilog),
        )

        index_path = filepath.parent.joinpath("index.rst")
        with index_path.open("w", encoding="utf8") as f:
            f.write(u(index_str))

        # clear any existing build
        build_dir = filepath.parent.joinpath("build/html")
        if build_dir.exists():
            # >> rm -r build/html
            shutil.rmtree(str(build_dir))
        build_dir.mkdir(parents=True)

        # run sphinx
        exec_path = find_executable("sphinx-build")
        args = [exec_path, "-b", "html"]
        if self.nitpick:
            args.append("-n")
        args.extend([str(filepath.parent.absolute()), str(build_dir.absolute())])

        self.logger.info("running: " + " ".join(args))

        # this way overrides the logging
        # sphinx_build = find_entry_point("sphinx-build", "console_scripts",
        #                                 self.logger, "sphinx")

        def log_process_output(pipe):
            for line in iter(pipe.readline, b""):
                self.logger.info("{}".format(line.decode("utf-8").strip()))

        process = Popen(args, stdout=PIPE, stderr=STDOUT)
        with process.stdout:
            log_process_output(process.stdout)
        exitcode = process.wait()  # 0 means success

        if exitcode:
            self.logger.warning("sphinx-build exited with code: {}".format(exitcode))

        if self.open_in_browser and not exitcode:
            # get entry path
            entry_path = filepath.parent.joinpath("build/html")
            entry_path = entry_path.joinpath(
                os.path.splitext(filepath.name)[0] + ".html"
            )
            if entry_path.exists():
                #  2 opens the url in a new tab
                webbrowser.open(entry_path.as_uri(), new=2)
            else:
                self.handle_error("can't find {0} to open".format(entry_path), IOError)

        return stream, filepath, resources



import os
import logging
from typing import Union  # noqa: F401

from docutils import nodes  # noqa E501
from docutils.parsers import rst
from ipypublish.utils import handle_error
from ipypublish.sphinx.utils import import_sphinx
from ipypublish.convert.main import IpyPubMain


# TODO should inherit from sphinx.parsers.RSTParser
# https://www.sphinx-doc.org/en/master/extdev/parserapi.html
# however, sphinx is an optional dependency
class NBParser(rst.Parser):
    """Sphinx source parser for Jupyter notebooks.
    adapted from nbsphinx
    """

    supported = ("jupyter_notebook",)

    def __init__(self, *args, **kwargs):

        self.app = None
        self.config = None
        self.env = None

        try:
            sphinx = import_sphinx()

            class NotebookError(sphinx.errors.SphinxError):
                """Error during notebook parsing."""

                category = "Notebook error"

            self.error_nb = NotebookError
            self.error_config = sphinx.errors.ConfigError
            self.logger = sphinx.util.logging.getLogger("nbparser")

        except (ImportError, AttributeError):
            self.error_nb = IOError
            self.error_config = TypeError
            self.logger = logging.getLogger("nbparser")

        super(NBParser, self).__init__(*args, **kwargs)

    def set_application(self, app):
        """set_application will be called from Sphinx to set app
        and other instance variables

        Parameters
        ----------
        app: sphinx.application.Sphinx
            Sphinx application object
        """
        self.app = app
        self.config = app.config
        self.env = app.env

    def parse(self, inputstring, document):
        # type: (Union[str, list[str]], nodes.document) -> None
        """Parse text and generate a document tree."""

        # fix for when calling on readthedocs
        self.env = self.env or document.settings.env
        self.config = self.config or document.settings.env.config

        # get file for conversion
        filepath = self.env.doc2path(self.env.docname)
        filedir = os.path.dirname(filepath)
        self.logger.info("ipypublish: converting {}".format(filepath))

        config = {
            "IpyPubMain": {
                "conversion": self.config.ipysphinx_export_config,
                "plugin_folder_paths": self.config.ipysphinx_config_folders,
                "outpath": filedir,
                "folder_suffix": self.config.ipysphinx_folder_suffix,
                "log_to_stdout": False,
                "log_to_file": False,
                "default_pporder_kwargs": dict(clear_existing=False, dump_files=True),
            }
        }
        if self.config.ipysphinx_preconverters:
            # NB: jupytext is already a default for .Rmd
            config["IpyPubMain"][
                "pre_conversion_funcs"
            ] = self.config.ipysphinx_preconverters
        publish = IpyPubMain(config=config)
        outdata = publish(filepath)

        self.logger.info("ipypublish: successful conversion")

        # check we got back restructuredtext
        exporter = outdata["exporter"]
        if not exporter.output_mimetype == "text/restructuredtext":
            handle_error(
                "ipypublish: the output content is not of type "
                "text/restructuredtext: {}".format(exporter.output_mimetype),
                TypeError,
                self.logger,
            )

        # TODO document use of orphan
        if outdata["resources"].get("ipub", {}).get("orphan", False):
            rst.Parser.parse(self, ":orphan:", document)

        # parse a prolog
        if self.env.config.ipysphinx_prolog:
            prolog = exporter.environment.from_string(
                self.env.config.ipysphinx_prolog
            ).render(env=self.env)
            rst.Parser.parse(self, prolog, document)

        # parse the main body of the file
        rst.Parser.parse(self, outdata["stream"], document)

        # parse an epilog
        if self.env.config.ipysphinx_epilog:
            prolog = exporter.environment.from_string(
                self.env.config.ipysphinx_epilog
            ).render(env=self.env)
            rst.Parser.parse(self, prolog, document)

        # TODO is there a better way to parse data back from the parser?

        # record if the notebook contains ipywidgets
        if outdata["resources"].get("contains_ipywidgets", False):
            if not hasattr(self.env, "ipysphinx_widgets"):
                self.env.ipysphinx_widgets = set()
            self.env.ipysphinx_widgets.add(self.env.docname)

        # record that the document was created from a notebook
        if not hasattr(self.env, "ipysphinx_created_from_nb"):
            self.env.ipysphinx_created_from_nb = set()
        self.env.ipysphinx_created_from_nb.add(self.env.docname)



import pytest

from ipypublish.convert import nbmerge
from ipypublish.convert.config_manager import create_exporter_cls, str_to_jinja
from ipypublish.convert.main import dict_to_config, IpyPubMain


@pytest.mark.ipynb("basic_nb")
def test_nbexport_latex_empty(ipynb_app):
    template = str_to_jinja("", "template_name")
    config = dict_to_config({"LatexExporter.template_file": "template_name"})
    exporter_cls = create_exporter_cls("nbconvert.exporters.LatexExporter")
    nb, path = nbmerge.merge_notebooks(ipynb_app.input_file)
    exporter, body, resources = IpyPubMain().export_notebook(
        nb, exporter_cls, config, template
    )
    assert exporter.output_mimetype == "text/latex"
    assert body == ""


@pytest.mark.ipynb("basic_nb")
def test_nbexport_latex_mkdown1(ipynb_app):
    template = str_to_jinja(
        """
((* block markdowncell scoped *))
test123
((* endblock markdowncell *))
    """,
        "template_name",
    )
    config = dict_to_config({"LatexExporter.template_file": "template_name"})
    nb, path = nbmerge.merge_notebooks(ipynb_app.input_file)
    exporter_cls = create_exporter_cls("nbconvert.exporters.LatexExporter")
    exporter, body, resources = IpyPubMain().export_notebook(
        nb, exporter_cls, config, template
    )
    assert exporter.output_mimetype == "text/latex"
    assert body.strip() == "test123"


@pytest.mark.ipynb("basic_nb")
def test_nbexport_latex_mkdown2(ipynb_app):
    template = str_to_jinja(
        """
((*- extends 'display_priority.tplx' -*))
((* block markdowncell scoped *))
(((cell.source)))
((* endblock markdowncell *))
    """,
        "template_name",
    )
    config = dict_to_config({"LatexExporter.template_file": "template_name"})
    nb, path = nbmerge.merge_notebooks(ipynb_app.input_file)
    exporter_cls = create_exporter_cls("nbconvert.exporters.LatexExporter")
    exporter, body, resources = IpyPubMain().export_notebook(
        nb, exporter_cls, config, template
    )
    assert exporter.output_mimetype == "text/latex"

    assert body.strip() == "# a title\n\nsome text"


@pytest.mark.ipynb("basic_nb")
def test_nbexport_html_empty(ipynb_app):
    template = str_to_jinja("", "template_name")
    config = dict_to_config({"HTMLExporter.template_file": "template_name"})
    nb, path = nbmerge.merge_notebooks(ipynb_app.input_file)
    exporter_cls = create_exporter_cls("nbconvert.exporters.HTMLExporter")
    exporter, body, resources = IpyPubMain().export_notebook(
        nb, exporter_cls, config, template
    )
    assert exporter.output_mimetype == "text/html"

    assert body == ""


@pytest.mark.ipynb("basic_nb")
def test_nbexport_html_mkdown1(ipynb_app):
    template = str_to_jinja(
        """
{% block markdowncell scoped %}
test123
{% endblock markdowncell %}
    """,
        "template_name",
    )
    config = dict_to_config({"HTMLExporter.template_file": "template_name"})
    nb, path = nbmerge.merge_notebooks(ipynb_app.input_file)
    exporter_cls = create_exporter_cls("nbconvert.exporters.HTMLExporter")
    exporter, body, resources = IpyPubMain().export_notebook(
        nb, exporter_cls, config, template
    )
    assert exporter.output_mimetype == "text/html"

    assert body.strip() == "test123"


@pytest.mark.ipynb("basic_nb")
def test_nbexport_html_mkdown2(ipynb_app):
    template = str_to_jinja(
        """
{%- extends 'display_priority.tpl' -%}
{% block markdowncell scoped %}
{{cell.source}}
{% endblock markdowncell %}
    """,
        "template_name",
    )
    config = dict_to_config({"HTMLExporter.template_file": "template_name"})
    nb, path = nbmerge.merge_notebooks(ipynb_app.input_file)
    exporter_cls = create_exporter_cls("nbconvert.exporters.HTMLExporter")
    exporter, body, resources = IpyPubMain().export_notebook(
        nb, exporter_cls, config, template
    )
    assert exporter.output_mimetype == "text/html"

    assert body.strip() == "# a title\n\nsome text"



"""in this pytest conftest file we define an ipypublish test app fixture

This fixture supplies a configurable IpubTestApp, that can
run IPyPubMain on a specified source folder,
in a temporary folder (which will be deleted on teardown),
and has methods to test the output.

Usage:

.. code-block:: python

    @pytest.mark.ipynb('folder_name')
    def test_example(ipynb_app):
        ipynb_app.run({"conversion": "latex_ipypublish_main"})
        ipynb_app.assert_converted_exists()
        ipynb_app.assert_converted_contains(["regex1", "regex2"])
        ipynb_app.assert_converted_equals_expected("file_name")

@pytest.mark.ipynb accepts the key-word arguments:

- root: the root folder containing the test folders
  (default = ipypublish.tests.TEST_FILES_DIR))
- source: the folder name within <root>/<folder_name>
  that contains the source files (default = 'source')
- main_file: the file within <root>/<folder_name>/<source> to be converted
  (default = 'main.ipynb'). If None is given the source folder is used.
- converted: the folder name to output converted files to,
  either in the temporary folder (if out_to_temp=True) or
  <root>/<folder_name>/<converted> (if out_to_temp=False)
  (default = 'converted')
- expected: the folder name within <root>/<folder_name>
  that contains the expected output files (default = 'expected')
- out_to_temp: if True, converted files output to a temporary folder,
  that will be removed on teardown,
  otherwise output to <root>/<folder_name>/<converted> (will not be removed)
  (default=True)

"""
from collections import namedtuple
import copy
from difflib import context_diff
import io
import logging
import os
import shutil
import re
import tempfile

from nbconvert.utils.pandoc import get_pandoc_version
import pytest

from ipypublish.utils import pathlib
from ipypublish.tests import TEST_FILES_DIR
from ipypublish.convert.main import IpyPubMain

logger = logging.getLogger(__name__)


@pytest.fixture(autouse=True)
def dont_open_webbrowser(monkeypatch):
    def nullfunc(*arg, **kwrgs):
        pass

    monkeypatch.setattr("webbrowser.open", nullfunc)


@pytest.fixture
def external_export_plugin():
    return pathlib.Path(os.path.join(TEST_FILES_DIR, "example_new_plugin.json"))


@pytest.fixture
def temp_folder():
    out_folder = tempfile.mkdtemp()
    yield out_folder
    shutil.rmtree(out_folder)


@pytest.fixture
def ipynb_params(request):
    """
    parameters that is specified by '@pytest.mark.ipynb'
    for ipynb conversion tests
    """

    # ##### process pytest.mark.ipynb

    if hasattr(request.node, "iter_markers"):  # pytest-3.6.0 or newer
        markers = request.node.iter_markers("ipynb")
    else:
        markers = request.node.get_marker("ipynb")
    pargs = {}
    kwargs = {}

    if markers is not None:
        # to avoid stacking positional args
        for info in reversed(list(markers)):
            for i, a in enumerate(info.args):
                pargs[i] = a
            kwargs.update(info.kwargs)

    args = [pargs[i] for i in sorted(pargs.keys())]

    return namedtuple("ipynb_params", "args,kwargs")(args, kwargs)  # type: ignore


@pytest.fixture(scope="function")
def ipynb_app(temp_folder, ipynb_params):

    args, kwargs = ipynb_params
    if len(args) <= 0:
        raise ValueError(
            "a subfolder must be supplied as the first argument to "
            "@pytest.mark.ipynb"
        )

    subfolder = args[0]  # 'ipynb_with_glossary'
    input_file = kwargs.get("main_file", "main.ipynb")
    test_files_dir = kwargs.get("root", TEST_FILES_DIR)
    source_folder = kwargs.get("source", "source")
    convert_folder = kwargs.get("converted", "converted")
    expected_folder = kwargs.get("expected", "expected")
    use_temp = kwargs.get("out_to_temp", True)

    source_folder_path = os.path.join(test_files_dir, subfolder, source_folder)
    expected_folder_path = os.path.join(test_files_dir, subfolder, expected_folder)

    temp_source_path = os.path.join(temp_folder, source_folder)
    shutil.copytree(source_folder_path, temp_source_path)

    if use_temp:
        converted_path = os.path.join(temp_folder, convert_folder)
    else:
        converted_path = os.path.join(test_files_dir, subfolder, convert_folder)

    yield IpyTestApp(temp_source_path, input_file, converted_path, expected_folder_path)


class IpyTestApp(object):
    def __init__(self, src_path, input_file, converted_path, expected_folder_path):
        self._src_folder_path = src_path
        self._converted_folder_path = converted_path
        self._expected_folder_path = expected_folder_path
        self._input_file = input_file
        self._output_data = None

    @property
    def source_path(self):
        return pathlib.Path(self._src_folder_path)

    @property
    def input_file(self):
        if self._input_file is None:
            return None
        return self.source_path.joinpath(self._input_file)

    @property
    def converted_path(self):
        return pathlib.Path(self._converted_folder_path)

    @property
    def expected_path(self):
        return pathlib.Path(self._expected_folder_path)

    @property
    def pandoc_version(self):
        return get_pandoc_version()

    def run(self, ipub_config=None):
        if ipub_config is None:
            ipub_config = {}
        ipub_config["outpath"] = str(self.converted_path)
        app = IpyPubMain(config={"IpyPubMain": ipub_config})
        self._output_data = app(
            self.input_file if self.input_file is not None else self.source_path
        )

    @property
    def output_data(self):
        if self._output_data is None:
            raise ValueError("the app must be run first to retrieve output data")
        return copy.copy(self._output_data)

    @property
    def export_extension(self):
        if self._output_data is None:
            raise ValueError(
                "the app must be run first to retrieve export file extension"
            )
        exporter = self._output_data["exporter"]
        return exporter.file_extension

    @property
    def export_mimetype(self):
        if self._output_data is None:
            raise ValueError("the app must be run first to retrieve export mimetype")
        exporter = self._output_data["exporter"]
        return exporter.output_mimetype

    def assert_converted_exists(self, file_name=None, extension=None):
        if file_name is None:
            if self.input_file is None:
                file_name = self.source_path.name
            else:
                file_name = os.path.splitext(self.input_file.name)[0]
        if extension is None:
            extension = self.export_extension

        converted_path = self.converted_path.joinpath(file_name + extension)
        if not self.converted_path.joinpath(file_name + extension).exists():
            raise AssertionError("could not find: {}".format(converted_path))

    def assert_converted_contains(self, regexes, encoding="utf8"):

        if self.input_file is None:
            file_name = self.source_path.name
        else:
            file_name = os.path.splitext(self.input_file.name)[0]
        extension = self.export_extension
        converted_path = self.converted_path.joinpath(file_name + extension)

        with io.open(str(converted_path), encoding=encoding) as fobj:
            content = fobj.read()

        if not isinstance(regexes, (list, tuple)):
            regexes = [regexes]

        for regex in regexes:

            if not re.search(regex, content):
                raise AssertionError("content does not contain regex: {}".format(regex))

    def assert_converted_equals_expected(self, expected_file_name, encoding="utf8"):

        if self.input_file is None:
            file_name = self.source_path.name
        else:
            file_name = os.path.splitext(self.input_file.name)[0]
        extension = self.export_extension
        converted_path = self.converted_path.joinpath(file_name + extension)

        expected_path = self.expected_path.joinpath(expected_file_name + extension)

        mime_type = self.export_mimetype
        if mime_type == "text/latex":
            compare_tex_files(converted_path, expected_path, encoding=encoding)
        elif mime_type == "text/html":
            compare_html_files(converted_path, expected_path, encoding=encoding)
        elif mime_type == "text/restructuredtext":
            compare_rst_files(converted_path, expected_path, encoding=encoding)
        else:
            # TODO add comparison for nb (applicatio/json)
            # and python (application/x-python)
            message = "no comparison function exists for " "mimetype: {}".format(
                mime_type
            )
            # raise ValueError(message)
            logger.warning(message)


def compare_rst_files(testpath, outpath, encoding="utf8"):
    # only compare body of html, since styles differ by
    # nbconvert/pandoc version (e.g. different versions of font-awesome)

    output = []
    for path in [testpath, outpath]:

        with io.open(str(path), encoding=encoding) as fobj:
            content = fobj.read()

        # python 3.5 used .jpg instead of .jpeg
        content = content.replace(".jpg", ".jpeg")

        # a recent dependency change is inserting new lines at the top of the file
        content = content.lstrip()

        output.append(content)

    test_content, out_content = output

    # only report differences
    if out_content != test_content:
        raise AssertionError(
            "\n"
            + "\n".join(
                context_diff(
                    test_content.splitlines(),
                    out_content.splitlines(),
                    fromfile=str(testpath),
                    tofile=str(outpath),
                )
            )
        )


def compare_html_files(testpath, outpath, encoding="utf8"):
    # only compare body of html, since styles differ by
    # nbconvert/pandoc version (e.g. different versions of font-awesome)

    output = []
    for path in [testpath, outpath]:

        with io.open(str(path), encoding=encoding) as fobj:
            content = fobj.read()

        # extract only the body
        # could use html.parser or beautifulsoup to do this better
        body_rgx = re.compile("\\<body\\>(.*)\\</body\\>", re.DOTALL)
        body_search = body_rgx.search(content)
        if not body_search:
            raise IOError("could not find body content of {}".format(path))
        content = body_search.group(1)

        # remove script environments which can change (e.g. reveal)
        script_rgx = re.compile("\\<script\\>(.*)\\</script\\>", re.DOTALL)
        content = script_rgx.sub("<script></script>", content)

        # remove trailing whitespace
        content = "\n".join([l.rstrip() for l in content.splitlines()])

        output.append(content)

    test_content, out_content = output

    # only report differences
    if out_content != test_content:
        raise AssertionError(
            "\n"
            + "\n".join(
                context_diff(
                    test_content.splitlines(),
                    out_content.splitlines(),
                    fromfile=str(testpath),
                    tofile=str(outpath),
                )
            )
        )


def compare_tex_files(testpath, outpath, encoding="utf8"):

    output = []
    for path in [testpath, outpath]:

        with io.open(str(path), encoding=encoding) as fobj:
            content = fobj.read()

        # only certain versions of pandoc wrap sections with \hypertarget
        # NOTE a better way to do this might be to use TexSoup
        ht_rgx = re.compile(
            "\\\\hypertarget\\{[^\\}]*\\}\\{[^\\\\]*"
            "(\\\\[sub]*section\\{[^\\}]*\\}"
            "\\\\label\\{[^\\}]*\\})"
            "\\}",
            re.DOTALL,
        )
        content = ht_rgx.sub("\\g<1>", content)

        # newer versions of pandoc convert ![](file) to \begin{figure}[htbp]
        # TODO override pandoc figure placement of ![](file) in markdown2latex
        content = content.replace("\\begin{figure}[htbp]", "\\begin{figure}")

        # at start of itemize
        content = content.replace("\\itemsep1pt\\parskip0pt\\parsep0pt\n", "")
        # at start of enumerate
        content = content.replace("\\tightlist\n", "")

        # python 3.5 used .jpg instead of .jpeg
        content = content.replace(".jpg", ".jpeg")

        # python < 3.6 sorts these differently
        pyg_rgx = re.compile(
            (
                "\\\\expandafter\\\\def\\\\csname "
                "PY\\@tok\\@[0-9a-zA-Z]*\\\\endcsname[^\n]*"
            ),
            re.MULTILINE,
        )
        content = pyg_rgx.sub(r"\<pygments definition\>", content)

        # also remove all space from start of lines
        space_rgx = re.compile(r"^[\s]*", re.MULTILINE)
        content = space_rgx.sub("", content)

        # remove trailing whitespace
        content = "\n".join([l.rstrip() for l in content.splitlines()])

        output.append(content)

    test_content, out_content = output

    # only report differences
    if out_content != test_content:
        raise AssertionError(
            "\n"
            + "\n".join(
                context_diff(
                    test_content.splitlines(),
                    out_content.splitlines(),
                    fromfile=str(testpath),
                    tofile=str(outpath),
                )
            )
        )



#!/usr/bin/env python
# import base64
from contextlib import contextmanager
from typing import List, Tuple, Union, Dict  # noqa: F401
import logging
import os
import time
import sys
import inspect

import traitlets as T
from traitlets import default, validate, TraitError

# from traitlets import validate
from traitlets.config.configurable import Configurable
from traitlets.config import Config
from jsonextended import edict
from six import string_types
import jsonschema

import ipypublish
from ipypublish.utils import (
    pathlib,
    handle_error,
    read_file_from_directory,
    get_module_path,
    get_valid_filename,
    find_entry_point,
)
from ipypublish import schema
from ipypublish.convert.nbmerge import merge_notebooks
from ipypublish.convert.config_manager import (
    get_export_config_path,
    load_export_config,
    load_template,
    create_exporter_cls,
)


def dict_to_config(config, unflatten=True, key_as_tuple=False):
    if unflatten:
        config = edict.unflatten(config, key_as_tuple=key_as_tuple, delim=".")
    return Config(config)


class IpyPubMain(Configurable):

    conversion = T.Unicode(
        "latex_ipypublish_main", help="key or path to conversion configuration"
    ).tag(config=True)

    plugin_folder_paths = T.Set(
        T.Unicode(),
        default_value=(),
        help="a list of folders containing conversion configurations",
    ).tag(config=True)

    @validate("plugin_folder_paths")
    def _validate_plugin_folder_paths(self, proposal):
        folder_paths = proposal["value"]
        for path in folder_paths:
            if not os.path.exists(path):
                raise TraitError(
                    "the configuration folder path does not exist: " "{}".format(path)
                )
        return proposal["value"]

    outpath = T.Union(
        [T.Unicode(), T.Instance(pathlib.Path)],
        allow_none=True,
        default_value=None,
        help="path to output converted files",
    ).tag(config=True)

    folder_suffix = T.Unicode(
        "_files",
        help=(
            "suffix for the folder name where content will be dumped "
            "(e.g. internal images). "
            "It will be a sanitized version of the input filename, "
            "followed by the suffix"
        ),
    ).tag(config=True)

    ignore_prefix = T.Unicode(
        "_", help=("prefixes to ignore, " "when finding notebooks to merge")
    ).tag(config=True)

    meta_path_placeholder = T.Unicode(
        "${meta_path}",
        help=(
            "all string values in the export configuration containing "
            "this placeholder will be be replaced with the path to the "
            "notebook from which the metadata was obtained"
        ),
    ).tag(config=True)

    files_folder_placeholder = T.Unicode(
        "${files_path}",
        help=(
            "all string values in the export configuration containing "
            "this placeholder will be be replaced with the path "
            "(relative to outpath) to the folder where files will be dumped"
        ),
    ).tag(config=True)

    validate_nb_metadata = T.Bool(
        True,
        help=(
            "before running the exporter, validate that "
            "the notebook level metadata is valid again the schema"
        ),
    ).tag(config=True)

    pre_conversion_funcs = T.Dict(
        help=(
            "a mapping of file extensions to functions that can convert"
            "that file type Instance(nbformat.NotebookNode) = func(pathstr)"
        )
    ).tag(config=True)

    @default("pre_conversion_funcs")
    def _default_pre_conversion_funcs(self):
        try:
            import jupytext  # noqa: F401
        except ImportError:
            return {}

        try:
            from jupytext import read
        except ImportError:
            # this is deprecated in newer versions
            from jupytext import readf as read  # noqa: F401

        return {".Rmd": read}

    @validate("pre_conversion_funcs")
    def _validate_pre_conversion_funcs(self, proposal):
        for ext, func in proposal["value"].items():
            if not ext.startswith("."):
                raise TraitError(
                    "the extension key should start with a '.': " "{}".format(ext)
                )
            try:
                func("string")
                # TODO should do this safely with inspect,
                # but no obvious solution
                # to check if it only requires one string argument
            except TypeError:
                raise TraitError(
                    "the function for {} can not be "
                    "called with a single string arg: "
                    "{}".format(ext, func)
                )
            except Exception:
                pass
        return proposal["value"]

    log_to_stdout = T.Bool(True, help="whether to log to sys.stdout").tag(config=True)

    log_level_stdout = T.Enum(
        ["debug", "info", "warning", "error", "DEBUG", "INFO", "WARNING", "ERROR"],
        default_value="INFO",
        help="the logging level to output to stdout",
    ).tag(config=True)

    log_stdout_formatstr = T.Unicode("%(levelname)s:%(name)s:%(message)s").tag(
        config=True
    )

    log_to_file = T.Bool(False, help="whether to log to file").tag(config=True)

    log_level_file = T.Enum(
        ["debug", "info", "warning", "error", "DEBUG", "INFO", "WARNING", "ERROR"],
        default_value="INFO",
        help="the logging level to output to file",
    ).tag(config=True)

    log_file_path = T.Unicode(
        None,
        allow_none=True,
        help="if None, will output to {outdir}/{ipynb_name}.nbpub.log",
    ).tag(config=True)

    log_file_formatstr = T.Unicode("%(levelname)s:%(name)s:%(message)s").tag(
        config=True
    )

    default_ppconfig_kwargs = T.Dict(
        trait=T.Bool(),
        default_value=(
            ("pdf_in_temp", False),
            ("pdf_debug", False),
            ("launch_browser", False),
        ),
        help=(
            "convenience arguments for constructing the post-processors "
            "default configuration"
        ),
    ).tag(config=True)

    default_pporder_kwargs = T.Dict(
        trait=T.Bool(),
        default_value=(
            ("dry_run", False),
            ("clear_existing", False),
            ("dump_files", False),
            ("create_pdf", False),
            ("serve_html", False),
            ("slides", False),
        ),
        help=(
            "convenience arguments for constructing the post-processors " "default list"
        ),
    ).tag(config=True)

    # TODO validate that default_ppconfig/pporder_kwargs can be parsed to funcs

    default_exporter_config = T.Dict(help="default configuration for exporters").tag(
        config=True
    )

    @default("default_exporter_config")
    def _default_exporter_config(self):
        temp = "${files_path}/{unique_key}_{cell_index}_{index}{extension}"
        return {"ExtractOutputPreprocessor": {"output_filename_template": temp}}

    def _create_default_ppconfig(
        self, pdf_in_temp=False, pdf_debug=False, launch_browser=False
    ):
        """create a default config for postprocessors"""
        return Config(
            {
                "PDFExport": {
                    "files_folder": "${files_path}",
                    "convert_in_temp": pdf_in_temp,
                    "debug_mode": pdf_debug,
                    "open_in_browser": launch_browser,
                    "skip_mime": False,
                },
                "RunSphinx": {"open_in_browser": launch_browser},
                "RemoveFolder": {"files_folder": "${files_path}"},
                "CopyResourcePaths": {"files_folder": "${files_path}"},
                "ConvertBibGloss": {"files_folder": "${files_path}"},
            }
        )

    def _create_default_pporder(
        self,
        dry_run=False,
        clear_existing=False,
        dump_files=False,
        create_pdf=False,
        serve_html=False,
        slides=False,
    ):
        """create a default list of postprocessors to run"""
        default_pprocs = [
            "remove-blank-lines",
            "remove-trailing-space",
            "filter-output-files",
        ]
        if slides:
            default_pprocs.append("fix-slide-refs")
        if not dry_run:
            if clear_existing:
                default_pprocs.append("remove-folder")
            default_pprocs.append("write-text-file")
            if dump_files or create_pdf or serve_html:
                default_pprocs.extend(
                    ["write-resource-files", "copy-resource-paths", "convert-bibgloss"]
                )
            if create_pdf:
                default_pprocs.append("pdf-export")
            elif serve_html:
                default_pprocs.append("reveal-server")

        return default_pprocs

    @property
    def logger(self):
        return logging.getLogger("ipypublish")

    @contextmanager
    def _log_handlers(self, ipynb_name, outdir):

        root = logging.getLogger()
        root_level = root.level
        log_handlers = []

        try:
            root.setLevel(logging.DEBUG)

            if self.log_to_stdout:
                # setup logging to terminal
                slogger = logging.StreamHandler(sys.stdout)
                slogger.setLevel(getattr(logging, self.log_level_stdout.upper()))
                formatter = logging.Formatter(self.log_stdout_formatstr)
                slogger.setFormatter(formatter)
                slogger.propogate = False
                root.addHandler(slogger)
                log_handlers.append(slogger)

            if self.log_to_file:
                # setup logging to file
                if self.log_file_path:
                    path = self.log_file_path
                else:
                    path = os.path.join(outdir, ipynb_name + ".nbpub.log")

                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path))

                flogger = logging.FileHandler(path, "w")
                flogger.setLevel(getattr(logging, self.log_level_file.upper()))
                formatter = logging.Formatter(self.log_file_formatstr)
                flogger.setFormatter(formatter)
                flogger.propogate = False
                root.addHandler(flogger)
                log_handlers.append(flogger)

            yield

        finally:

            root.setLevel(root_level)
            for handler in log_handlers:
                handler.close()
                root.removeHandler(handler)

    def __init__(self, config=None):
        """
        Public constructor

        Parameters
        ----------
        config: traitlets.config.Config
            User configuration instance.

        """
        # with_default_config = self.default_config
        # if config:
        #     with_default_config.merge(config)
        if config is None:
            config = {}
        if not isinstance(config, Config):
            config = Config(config)
        with_default_config = config

        super(IpyPubMain, self).__init__(config=with_default_config)

    def __call__(self, ipynb_path, nb_node=None):
        """see IpyPubMain.publish"""
        return self.publish(ipynb_path, nb_node)

    def publish(self, ipynb_path, nb_node=None):
        """ convert one or more Jupyter notebooks to a published format

        paths can be string of an existing file or folder,
        or a pathlib.Path like object

        all files linked in the documents are placed into a single files_folder

        Parameters
        ----------
        ipynb_path: str or pathlib.Path
            notebook file or directory
        nb_node: None or nbformat.NotebookNode
            a pre-converted notebook

        Returns
        --------
        outdata: dict
            containing keys;
            "outpath", "exporter", "stream", "main_filepath", "resources"

        """
        # setup the input and output paths
        if isinstance(ipynb_path, string_types):
            ipynb_path = pathlib.Path(ipynb_path)
        ipynb_name, ipynb_ext = os.path.splitext(ipynb_path.name)
        outdir = (
            os.path.join(os.getcwd(), "converted")
            if self.outpath is None
            else str(self.outpath)
        )

        with self._log_handlers(ipynb_name, outdir):

            if not ipynb_path.exists() and not nb_node:
                handle_error(
                    "the notebook path does not exist: {}".format(ipynb_path),
                    IOError,
                    self.logger,
                )

            # log start of conversion
            self.logger.info(
                "started ipypublish v{0} at {1}".format(
                    ipypublish.__version__, time.strftime("%c")
                )
            )
            self.logger.info(
                "logging to: {}".format(os.path.join(outdir, ipynb_name + ".nbpub.log"))
            )
            self.logger.info("running for ipynb(s) at: {0}".format(ipynb_path))
            self.logger.info(
                "with conversion configuration: {0}".format(self.conversion)
            )

            if nb_node is None and ipynb_ext in self.pre_conversion_funcs:
                func = self.pre_conversion_funcs[ipynb_ext]
                self.logger.info(
                    "running pre-conversion with: {}".format(inspect.getmodule(func))
                )
                try:
                    nb_node = func(ipynb_path)
                except Exception as err:
                    handle_error(
                        "pre-conversion failed for {}: {}".format(ipynb_path, err),
                        err,
                        self.logger,
                    )

            # doesn't work with folders
            # if (ipynb_ext != ".ipynb" and nb_node is None):
            #     handle_error(
            #         'the file extension is not associated with any '
            #         'pre-converter: {}'.format(ipynb_ext),
            # TypeError, self.logger)

            if nb_node is None:
                # merge all notebooks
                # TODO allow notebooks to remain separate
                # (would require creating a main.tex with the preamble in etc )
                # Could make everything a 'PyProcess',
                # with support for multiple streams
                final_nb, meta_path = merge_notebooks(
                    ipynb_path, ignore_prefix=self.ignore_prefix
                )
            else:
                final_nb, meta_path = (nb_node, ipynb_path)

            # validate the notebook metadata against the schema
            if self.validate_nb_metadata:
                nb_metadata_schema = read_file_from_directory(
                    get_module_path(schema),
                    "doc_metadata.schema.json",
                    "doc_metadata.schema",
                    self.logger,
                    interp_ext=True,
                )
                try:
                    jsonschema.validate(final_nb.metadata, nb_metadata_schema)
                except jsonschema.ValidationError as err:
                    handle_error(
                        "validation of notebook level metadata failed: {}\n"
                        "see the doc_metadata.schema.json for full spec".format(
                            err.message
                        ),
                        jsonschema.ValidationError,
                        logger=self.logger,
                    )

            # set text replacements for export configuration
            replacements = {
                self.meta_path_placeholder: str(meta_path),
                self.files_folder_placeholder: "{}{}".format(
                    get_valid_filename(ipynb_name), self.folder_suffix
                ),
            }

            self.logger.debug("notebooks meta path: {}".format(meta_path))

            # load configuration file
            (
                exporter_cls,
                jinja_template,
                econfig,
                pprocs,
                pconfig,
            ) = self._load_config_file(replacements)

            # run nbconvert
            self.logger.info("running nbconvert")
            exporter, stream, resources = self.export_notebook(
                final_nb, exporter_cls, econfig, jinja_template
            )

            # postprocess results
            main_filepath = os.path.join(outdir, ipynb_name + exporter.file_extension)

            for post_proc_name in pprocs:
                proc_class = find_entry_point(
                    post_proc_name,
                    "ipypublish.postprocessors",
                    self.logger,
                    "ipypublish",
                )
                proc = proc_class(pconfig)
                stream, main_filepath, resources = proc.postprocess(
                    stream, exporter.output_mimetype, main_filepath, resources
                )

            self.logger.info("process finished successfully")

        return {
            "outpath": outdir,
            "exporter": exporter,
            "stream": stream,
            "main_filepath": main_filepath,
            "resources": resources,
        }

    def _load_config_file(self, replacements):
        # find conversion configuration
        self.logger.info("finding conversion configuration: {}".format(self.conversion))
        export_config_path = None
        if isinstance(self.conversion, string_types):
            outformat_path = pathlib.Path(self.conversion)
        else:
            outformat_path = self.conversion
        if outformat_path.exists():  # TODO use pathlib approach
            # if is outformat is a path that exists, use that
            export_config_path = outformat_path
        else:
            # else search internally
            export_config_path = get_export_config_path(
                self.conversion, self.plugin_folder_paths
            )

        if export_config_path is None:
            handle_error(
                "could not find conversion configuration: {}".format(self.conversion),
                IOError,
                self.logger,
            )

        # read conversion configuration and create
        self.logger.info("loading conversion configuration")
        data = load_export_config(export_config_path)
        self.logger.info("creating exporter")
        exporter_cls = create_exporter_cls(data["exporter"]["class"])
        self.logger.info("creating template and loading filters")
        template_name = "template_file"
        jinja_template = load_template(template_name, data["template"])
        self.logger.info("creating process configuration")
        export_config = self._create_export_config(
            data["exporter"], template_name, replacements
        )
        pprocs, pproc_config = self._create_pproc_config(
            data.get("postprocessors", {}), replacements
        )

        return (exporter_cls, jinja_template, export_config, pprocs, pproc_config)

    def _create_export_config(self, exporter_data, template_name, replacements):
        # type: (dict, Dict[str, str]) -> Config
        config = {}
        exporter_name = exporter_data["class"].split(".")[-1]

        config[exporter_name + ".template_file"] = template_name
        config[exporter_name + ".filters"] = exporter_data.get("filters", [])

        preprocessors = []
        for preproc in exporter_data.get("preprocessors", []):
            preprocessors.append(preproc["class"])
            preproc_name = preproc["class"].split(".")[-1]
            for name, val in preproc.get("args", {}).items():
                config[preproc_name + "." + name] = val

        config[exporter_name + ".preprocessors"] = preprocessors

        for name, val in exporter_data.get("other_args", {}).items():
            config[name] = val

        final_config = self.default_exporter_config
        final_config.update(config)

        replace_placeholders(final_config, replacements)

        return dict_to_config(final_config, True)

    def _create_pproc_config(self, pproc_data, replacements):

        if "order" in pproc_data:
            pprocs_list = pproc_data["order"]
        else:
            pprocs_list = self._create_default_pporder(**self.default_pporder_kwargs)

        pproc_config = self._create_default_ppconfig(**self.default_ppconfig_kwargs)

        if "config" in pproc_data:
            override_config = pproc_data["config"]
            pproc_config.update(override_config)

        replace_placeholders(pproc_config, replacements)

        return pprocs_list, pproc_config

    def export_notebook(self, final_nb, exporter_cls, config, jinja_template):

        kwargs = {"config": config}
        if jinja_template is not None:
            kwargs["extra_loaders"] = [jinja_template]
        try:
            exporter = exporter_cls(**kwargs)
        except TypeError:
            self.logger.warning(
                "the exporter class can not be parsed "
                "the arguments: {}".format(list(kwargs.keys()))
            )
            exporter = exporter_cls()

        body, resources = exporter.from_notebook_node(final_nb)
        return exporter, body, resources


def replace_placeholders(mapping, replacements):
    """ recurse through a mapping and perform (in-place) string replacements

    Parameters
    ----------
    mapping:
        any object which has an items() attribute
    replacements: dict
        {placeholder: replacement}

    """
    for key, val in mapping.items():
        if isinstance(val, string_types):
            for instr, outstr in replacements.items():
                val = val.replace(instr, outstr)
            mapping[key] = val
        elif hasattr(val, "items"):
            replace_placeholders(val, replacements)



import shutil
import tempfile
import os

from ipypublish.convert.config_manager import iter_all_export_paths
from ipypublish.convert.main import IpyPubMain
from ipypublish.tests import TEST_FILES_DIR


def convert_all(inpath, outpath):
    """ convert notebook using all available plugins """
    for plugin_name, plugin_path in iter_all_export_paths():

        out_folder = tempfile.mkdtemp()
        publish = IpyPubMain(
            config={"IpyPubMain": {"conversion": plugin_name, "outpath": out_folder}}
        )
        try:
            outdata = publish(str(inpath))

            exporter = outdata["exporter"]
            outpath = outdata["outpath"]

            extension = exporter.file_extension
            out_name = os.path.splitext(os.path.basename(str(inpath)))[0] + extension
            outfile = os.path.join(out_folder, out_name)

            if not os.path.exists(outfile):
                raise IOError("could not find: {} for {}".format(outfile, plugin_name))

            shutil.copyfile(
                outfile, os.path.join(str(outpath), plugin_name + extension)
            )

        finally:
            shutil.rmtree(out_folder)


if __name__ == "__main__":

    _inpath = os.path.join(TEST_FILES_DIR, "ipynb1.ipynb")
    _outpath = os.path.join(TEST_FILES_DIR, "ipynb1_converted")
    convert_all(_inpath, _outpath)



"""
Uses sphinx's pytest fixture to run builds

usage:

.. code-block:: python

    from ipypublish.sphinx.tests import get_test_source_dir

    @pytest.mark.sphinx(
        buildername='html',
        srcdir=get_test_source_dir('notebook'))
    def test_basic(app, status, warning, get_sphinx_app_output):

        app.build()

        assert 'build succeeded' in status.getvalue()  # Build succeeded
        warnings = warning.getvalue().strip()
        assert warnings == ""

        output = get_sphinx_app_output(app, buildername='html')

parameters available to parse to ``@pytest.mark.sphinx``:

- buildername='html'
- srcdir=None
- testroot='root' (only used if srcdir not set)
- freshenv=False
- confoverrides=None
- status=None
- warning=None
- tags=None
- docutilsconf=None

"""
import os
import shutil
import re

import pytest
from sphinx.testing.path import path

try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

from ipypublish.sphinx.tests import get_test_source_dir


@pytest.fixture(scope="session", autouse=True)
def remove_sphinx_builds():
    """ remove all build directories from the test folder
    """
    srcdirs = pathlib.Path(get_test_source_dir())
    for entry in srcdirs.iterdir():  # type: pathlib.Path
        if entry.is_dir() and entry.joinpath("_build").exists():
            shutil.rmtree(str(entry.joinpath("_build")))


@pytest.fixture
def get_sphinx_app_output():
    def read(
        app,
        buildername="html",
        filename="contents.html",
        encoding="utf-8",
        extract_body=False,
        remove_scripts=False,
    ):

        outpath = path(os.path.join(str(app.srcdir), "_build", buildername, filename))
        if not outpath.exists():
            raise IOError("no output file exists: {}".format(outpath))

        content = outpath.text(encoding=encoding)

        if extract_body:
            body_rgx = re.compile("\\<body\\>(.*)\\</body\\>", re.DOTALL)
            body_search = body_rgx.search(content)
            if not body_search:
                raise IOError("could not find body content of {}".format(path))
            content = body_search.group(1)

        if remove_scripts:
            # remove script environments which can change
            script_rgx = re.compile("\\<script\\>(.*)\\</script\\>", re.DOTALL)
            content = script_rgx.sub("<script></script>", content)

        return content

    return read
